'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _Map = _interopDefault(require('babel-runtime/core-js/map'));
var _regeneratorRuntime = _interopDefault(require('babel-runtime/regenerator'));
var _asyncToGenerator = _interopDefault(require('babel-runtime/helpers/asyncToGenerator'));
var _Object$getPrototypeOf = _interopDefault(require('babel-runtime/core-js/object/get-prototype-of'));
var _classCallCheck = _interopDefault(require('babel-runtime/helpers/classCallCheck'));
var _createClass = _interopDefault(require('babel-runtime/helpers/createClass'));
var _possibleConstructorReturn = _interopDefault(require('babel-runtime/helpers/possibleConstructorReturn'));
var _inherits = _interopDefault(require('babel-runtime/helpers/inherits'));
var _getIterator = _interopDefault(require('babel-runtime/core-js/get-iterator'));
var path = _interopDefault(require('path'));
var pify = _interopDefault(require('pify'));
var _extends = _interopDefault(require('babel-runtime/helpers/extends'));
var _typeof = _interopDefault(require('babel-runtime/helpers/typeof'));
var _slicedToArray = _interopDefault(require('babel-runtime/helpers/slicedToArray'));
var _Math$sign = _interopDefault(require('babel-runtime/core-js/math/sign'));
var buffer = require('buffer');
var openpgp = require('openpgp/dist/openpgp.min.js');
var _Object$keys = _interopDefault(require('babel-runtime/core-js/object/keys'));
var ini = _interopDefault(require('ini'));
var _get = _interopDefault(require('lodash/get'));
var _set = _interopDefault(require('lodash/set'));
var pako = _interopDefault(require('pako'));
var shasum = _interopDefault(require('shasum'));
var BufferCursor = _interopDefault(require('buffercursor'));
var pad = _interopDefault(require('pad'));
var gartal = require('gartal');
var _toConsumableArray = _interopDefault(require('babel-runtime/helpers/toConsumableArray'));
var _Object$assign = _interopDefault(require('babel-runtime/core-js/object/assign'));
var _Array$from = _interopDefault(require('babel-runtime/core-js/array/from'));
var _Symbol$iterator = _interopDefault(require('babel-runtime/core-js/symbol/iterator'));
var sortby = _interopDefault(require('lodash/sortBy'));
var _Promise = _interopDefault(require('babel-runtime/core-js/promise'));
var _Set = _interopDefault(require('babel-runtime/core-js/set'));
var AsyncLock = _interopDefault(require('async-lock'));
var ignore = _interopDefault(require('ignore'));
var simpleGet = _interopDefault(require('simple-get'));
var concat = _interopDefault(require('simple-concat'));
var stream = require('stream');
var stream__default = _interopDefault(stream);
var _Number$isNaN = _interopDefault(require('babel-runtime/core-js/number/is-nan'));
var thru = _interopDefault(require('thru'));
var listpack = _interopDefault(require('git-list-pack'));
var peek = _interopDefault(require('buffer-peek-stream'));
var applyDelta = _interopDefault(require('git-apply-delta'));
var marky = _interopDefault(require('marky'));
var crypto = _interopDefault(require('crypto'));

// @flow
function formatTimezoneOffset(minutes /*: number */) {
  var sign$$1 = _Math$sign(minutes) || 1;
  minutes = Math.abs(minutes);
  var hours = Math.floor(minutes / 60);
  minutes -= hours * 60;
  var strHours = String(hours);
  var strMinutes = String(minutes);
  if (strHours.length < 2) strHours = '0' + strHours;
  if (strMinutes.length < 2) strMinutes = '0' + strMinutes;
  return (sign$$1 === 1 ? '-' : '+') + strHours + strMinutes;
}

function parseTimezoneOffset(offset) {
  var _offset$match = offset.match(/(\+|-)(\d\d)(\d\d)/),
      _offset$match2 = _slicedToArray(_offset$match, 4),
      sign$$1 = _offset$match2[1],
      hours = _offset$match2[2],
      minutes = _offset$match2[3];

  minutes = (sign$$1 === '-' ? 1 : -1) * Number(hours) * 60 + Number(minutes);
  return minutes;
}

function parseAuthor(author) {
  var _author$match = author.match(/^(.*) <(.*)> (.*) (.*)$/),
      _author$match2 = _slicedToArray(_author$match, 5),
      name = _author$match2[1],
      email = _author$match2[2],
      timestamp = _author$match2[3],
      offset = _author$match2[4];

  return {
    name: name,
    email: email,
    timestamp: Number(timestamp),
    timezoneOffset: parseTimezoneOffset(offset)
  };
}

function normalize(str) {
  // remove all <CR>
  str = str.replace(/\r/g, '');
  // no extra newlines up front
  str = str.replace(/^\n+/, '');
  // and a single newline at the end
  str = str.replace(/\n+$/, '') + '\n';
  return str;
}

function indent(str) {
  return str.trim().split('\n').map(function (x) {
    return ' ' + x;
  }).join('\n') + '\n';
}

function outdent(str) {
  return str.split('\n').map(function (x) {
    return x.replace(/^ /, '');
  }).join('\n');
}

// TODO: Make all functions have static async signature?

var GitCommit = function () {
  /*::
  _commit : string
  */
  function GitCommit(commit /*: string|Buffer|Object */) {
    _classCallCheck(this, GitCommit);

    if (typeof commit === 'string') {
      this._commit = commit;
    } else if (buffer.Buffer.isBuffer(commit)) {
      this._commit = commit.toString('utf8');
    } else if ((typeof commit === 'undefined' ? 'undefined' : _typeof(commit)) === 'object') {
      this._commit = GitCommit.render(commit);
    } else {
      throw new Error('invalid type passed to GitCommit constructor');
    }
  }

  _createClass(GitCommit, [{
    key: 'toObject',
    value: function toObject() {
      return buffer.Buffer.from(this._commit, 'utf8');
    }

    // Todo: allow setting the headers and message

  }, {
    key: 'headers',
    value: function headers() {
      return this.parseHeaders();
    }

    // Todo: allow setting the headers and message

  }, {
    key: 'message',
    value: function message$$1() {
      return GitCommit.justMessage(this._commit);
    }
  }, {
    key: 'parse',
    value: function parse() {
      return _extends({ message: this.message() }, this.headers());
    }
  }, {
    key: 'parseHeaders',
    value: function parseHeaders() {
      var headers = GitCommit.justHeaders(this._commit).split('\n');
      var hs = [];
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = _getIterator(headers), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var h = _step.value;

          if (h[0] === ' ') {
            // combine with previous header (without space indent)
            hs[hs.length - 1] += '\n' + h.slice(1);
          } else {
            hs.push(h);
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      var obj = {};
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = _getIterator(hs), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var _h = _step2.value;

          var key$$1 = _h.slice(0, _h.indexOf(' '));
          var value = _h.slice(_h.indexOf(' ') + 1);
          obj[key$$1] = value;
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }

      obj.parent = obj.parent ? obj.parent.split(' ') : [];
      if (obj.author) {
        obj.author = parseAuthor(obj.author);
      }
      if (obj.committer) {
        obj.committer = parseAuthor(obj.committer);
      }
      return obj;
    }
  }, {
    key: 'render',
    value: function render() {
      return this._commit;
    }
  }, {
    key: 'withoutSignature',
    value: function withoutSignature() {
      var commit = normalize(this._commit);
      if (commit.indexOf('\ngpgsig') === -1) return commit;
      var headers = commit.slice(0, commit.indexOf('\ngpgsig'));
      var message$$1 = commit.slice(commit.indexOf('-----END PGP SIGNATURE-----\n') + '-----END PGP SIGNATURE-----\n'.length);
      return normalize(headers + '\n' + message$$1);
    }
  }, {
    key: 'isolateSignature',
    value: function isolateSignature() {
      var signature = this._commit.slice(this._commit.indexOf('-----BEGIN PGP SIGNATURE-----'), this._commit.indexOf('-----END PGP SIGNATURE-----') + '-----END PGP SIGNATURE-----'.length);
      return outdent(signature);
    }
  }, {
    key: 'sign',
    value: function () {
      var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(privateKeys /*: string */) {
        var commit, headers, message$$1, privKeyObj, _ref2, signature, signedCommit;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                commit = this.withoutSignature();
                headers = GitCommit.justHeaders(this._commit);
                message$$1 = GitCommit.justMessage(this._commit);
                privKeyObj = openpgp.key.readArmored(privateKeys).keys;
                _context.next = 6;
                return openpgp.sign({
                  data: openpgp.util.str2Uint8Array(commit),
                  privateKeys: privKeyObj,
                  detached: true,
                  armor: true
                });

              case 6:
                _ref2 = _context.sent;
                signature = _ref2.signature;

                // renormalize the line endings to the one true line-ending
                signature = normalize(signature);
                signedCommit = headers + '\n' + 'gpgsig' + indent(signature) + '\n' + message$$1;
                // return a new commit object

                return _context.abrupt('return', GitCommit.from(signedCommit));

              case 11:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function sign$$1(_x) {
        return _ref.apply(this, arguments);
      }

      return sign$$1;
    }()
  }, {
    key: 'listSigningKeys',
    value: function () {
      var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2() {
        var msg;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                msg = openpgp.message.readSignedContent(this.withoutSignature(), this.isolateSignature());
                return _context2.abrupt('return', msg.getSigningKeyIds().map(function (keyid) {
                  return keyid.toHex();
                }));

              case 2:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function listSigningKeys() {
        return _ref3.apply(this, arguments);
      }

      return listSigningKeys;
    }()
  }, {
    key: 'verify',
    value: function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(publicKeys /*: string */) {
        var pubKeyObj, msg, results, validity;
        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                pubKeyObj = openpgp.key.readArmored(publicKeys).keys;
                msg = openpgp.message.readSignedContent(this.withoutSignature(), this.isolateSignature());
                results = msg.verify(pubKeyObj);
                validity = results.reduce(function (a, b) {
                  return a.valid && b.valid;
                }, { valid: true });
                return _context3.abrupt('return', validity);

              case 5:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function verify(_x2) {
        return _ref4.apply(this, arguments);
      }

      return verify;
    }()
  }], [{
    key: 'fromPayloadSignature',
    value: function fromPayloadSignature(_ref5) {
      var payload = _ref5.payload,
          signature = _ref5.signature;

      var headers = GitCommit.justHeaders(payload);
      var message$$1 = GitCommit.justMessage(payload);
      var commit = normalize(headers + '\ngpgsig' + indent(signature) + '\n' + message$$1);
      return new GitCommit(commit);
    }
  }, {
    key: 'from',
    value: function from(commit) {
      return new GitCommit(commit);
    }
  }, {
    key: 'justMessage',
    value: function justMessage(commit) {
      return normalize(commit.slice(commit.indexOf('\n\n') + 2));
    }
  }, {
    key: 'justHeaders',
    value: function justHeaders(commit) {
      return commit.slice(0, commit.indexOf('\n\n'));
    }
  }, {
    key: 'renderHeaders',
    value: function renderHeaders(obj) {
      var headers = '';
      if (obj.tree) {
        headers += 'tree ' + obj.tree + '\n';
      } else {
        headers += 'tree 4b825dc642cb6eb9a060e54bf8d69288fbee4904\n'; // the null tree
      }
      if (obj.parent) {
        headers += 'parent';
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
          for (var _iterator3 = _getIterator(obj.parent), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            var p = _step3.value;

            headers += ' ' + p;
          }
        } catch (err) {
          _didIteratorError3 = true;
          _iteratorError3 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion3 && _iterator3.return) {
              _iterator3.return();
            }
          } finally {
            if (_didIteratorError3) {
              throw _iteratorError3;
            }
          }
        }

        headers += '\n';
      }
      var author = obj.author;
      headers += 'author ' + author.name + ' <' + author.email + '> ' + author.timestamp + ' ' + formatTimezoneOffset(author.timezoneOffset) + '\n';
      var committer = obj.committer || obj.author;
      headers += 'committer ' + committer.name + ' <' + committer.email + '> ' + committer.timestamp + ' ' + formatTimezoneOffset(committer.timezoneOffset) + '\n';
      if (obj.gpgsig) {
        headers += 'gpgsig' + indent(obj.gpgsig);
      }
      return headers;
    }
  }, {
    key: 'render',
    value: function render(obj) {
      return GitCommit.renderHeaders(obj) + '\n' + normalize(obj.message);
    }
  }]);

  return GitCommit;
}();

var complexKeys = ['remote', 'branch'];

var isComplexKey = function isComplexKey(key$$1) {
  return complexKeys.reduce(function (x, y) {
    return x || key$$1.startsWith(y);
  }, false);
};

var splitComplexKey = function splitComplexKey(key$$1) {
  return key$$1.split('"').map(function (x) {
    return x.trim();
  }).filter(function (x) {
    return x !== '';
  });
};

// Note: there are a LOT of edge cases that aren't covered (e.g. keys in sections that also
// have subsections, [include] directives, etc.
var GitConfig = function () {
  function GitConfig(text) {
    _classCallCheck(this, GitConfig);

    this.ini = ini.decode(text);
    // Some mangling to make it easier to work with (honestly)
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = _getIterator(_Object$keys(this.ini)), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var key$$1 = _step.value;

        if (isComplexKey(key$$1)) {
          var parts = splitComplexKey(key$$1);
          if (parts.length === 2) {
            // just to be cautious
            _set(this.ini, [parts[0], parts[1]], this.ini[key$$1]);
            delete this.ini[key$$1];
          }
        }
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }
  }

  _createClass(GitConfig, [{
    key: 'get',
    value: function () {
      var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(path$$1) {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                return _context.abrupt('return', _get(this.ini, path$$1));

              case 1:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function get(_x) {
        return _ref.apply(this, arguments);
      }

      return get;
    }()
  }, {
    key: 'set',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(path$$1, value) {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                return _context2.abrupt('return', _set(this.ini, path$$1, value));

              case 1:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function set(_x2, _x3) {
        return _ref2.apply(this, arguments);
      }

      return set;
    }()
  }, {
    key: 'toString',
    value: function toString() {
      // de-mangle complex keys
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = _getIterator(_Object$keys(this.ini)), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var key$$1 = _step2.value;

          if (isComplexKey(key$$1)) {
            var _iteratorNormalCompletion3 = true;
            var _didIteratorError3 = false;
            var _iteratorError3 = undefined;

            try {
              for (var _iterator3 = _getIterator(_Object$keys(this.ini[key$$1])), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                var childkey = _step3.value;

                var complexkey = key$$1 + ' "' + childkey + '"';
                this.ini[complexkey] = this.ini[key$$1][childkey];
                delete this.ini[key$$1][childkey];
              }
            } catch (err) {
              _didIteratorError3 = true;
              _iteratorError3 = err;
            } finally {
              try {
                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                  _iterator3.return();
                }
              } finally {
                if (_didIteratorError3) {
                  throw _iteratorError3;
                }
              }
            }

            delete this.ini[key$$1];
          }
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }

      var text = ini.encode(this.ini, { whitespace: true });
      return text;
    }
  }], [{
    key: 'from',
    value: function from(text) {
      return new GitConfig(text);
    }
  }]);

  return GitConfig;
}();

var GitObject = function () {
  function GitObject() {
    _classCallCheck(this, GitObject);
  }

  _createClass(GitObject, null, [{
    key: 'wrap',
    value: function wrap(_ref /*: {type: string, object: Buffer} */) {
      var type = _ref.type,
          object = _ref.object;

      var buffer$$1 = buffer.Buffer.concat([buffer.Buffer.from(type + ' '), buffer.Buffer.from(object.byteLength.toString()), buffer.Buffer.from([0]), buffer.Buffer.from(object)]);
      var oid = shasum(buffer$$1);
      return {
        oid: oid,
        file: buffer.Buffer.from(pako.deflate(buffer$$1))
      };
    }
  }, {
    key: 'unwrap',
    value: function unwrap(_ref2 /*: {oid: string, file: Buffer} */) {
      var oid = _ref2.oid,
          file = _ref2.file;

      var inflated = buffer.Buffer.from(pako.inflate(file));
      if (oid) {
        var sha = shasum(inflated);
        if (sha !== oid) {
          throw new Error('SHA check failed! Expected ' + oid + ', computed ' + sha);
        }
      }
      var s = inflated.indexOf(32); // first space
      var i = inflated.indexOf(0); // first null value
      var type = inflated.slice(0, s).toString('utf8'); // get type of object
      var length = inflated.slice(s + 1, i).toString('utf8'); // get type of object
      var actualLength = inflated.length - (i + 1);
      // verify length
      if (parseInt(length) !== actualLength) {
        throw new Error('Length mismatch: expected ' + length + ' bytes but got ' + actualLength + ' instead.');
      }
      return {
        type: type,
        object: buffer.Buffer.from(inflated.slice(i + 1))
      };
    }
  }]);

  return GitObject;
}();

// @flow
/**
pkt-line Format
---------------

Much (but not all) of the payload is described around pkt-lines.

A pkt-line is a variable length binary string.  The first four bytes
of the line, the pkt-len, indicates the total length of the line,
in hexadecimal.  The pkt-len includes the 4 bytes used to contain
the length's hexadecimal representation.

A pkt-line MAY contain binary data, so implementors MUST ensure
pkt-line parsing/formatting routines are 8-bit clean.

A non-binary line SHOULD BE terminated by an LF, which if present
MUST be included in the total length. Receivers MUST treat pkt-lines
with non-binary data the same whether or not they contain the trailing
LF (stripping the LF if present, and not complaining when it is
missing).

The maximum length of a pkt-line's data component is 65516 bytes.
Implementations MUST NOT send pkt-line whose length exceeds 65520
(65516 bytes of payload + 4 bytes of length data).

Implementations SHOULD NOT send an empty pkt-line ("0004").

A pkt-line with a length field of 0 ("0000"), called a flush-pkt,
is a special case and MUST be handled differently than an empty
pkt-line ("0004").

----
  pkt-line     =  data-pkt / flush-pkt

  data-pkt     =  pkt-len pkt-payload
  pkt-len      =  4*(HEXDIG)
  pkt-payload  =  (pkt-len - 4)*(OCTET)

  flush-pkt    = "0000"
----

Examples (as C-style strings):

----
  pkt-line          actual value
  ---------------------------------
  "0006a\n"         "a\n"
  "0005a"           "a"
  "000bfoobar\n"    "foobar\n"
  "0004"            ""
----
*/
// I'm really using this more as a namespace.
// There's not a lot of "state" in a pkt-line
var GitPktLine = function () {
  function GitPktLine() {
    _classCallCheck(this, GitPktLine);
  }

  _createClass(GitPktLine, null, [{
    key: 'flush',
    value: function flush() {
      return buffer.Buffer.from('0000', 'utf8');
    }
  }, {
    key: 'encode',
    value: function encode(line /*: string|Buffer */) /*: Buffer */{
      if (typeof line === 'string') {
        line = buffer.Buffer.from(line);
      }
      var length = line.length + 4;
      var hexlength = pad(4, length.toString(16), '0');
      return buffer.Buffer.concat([buffer.Buffer.from(hexlength, 'utf8'), line]);
    }
  }, {
    key: 'reader',
    value: function reader(buffer$$1 /*: Buffer */) {
      var buffercursor = new BufferCursor(buffer$$1);
      return function read() {
        if (buffercursor.eof()) return true;
        var length = parseInt(buffercursor.slice(4).toString('utf8'), 16);
        if (length === 0) return null;
        return buffercursor.slice(length - 4).buffer;
      };
    }
  }, {
    key: 'streamReader',
    value: function streamReader(stream$$1 /*: ReadableStream */) {
      return function () {
        var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee() {
          var hexlength, length, bytes;
          return _regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
              switch (_context.prev = _context.next) {
                case 0:
                  hexlength = void 0, length = void 0, bytes = void 0;
                  _context.prev = 1;
                  _context.next = 4;
                  return gartal.readBytes(stream$$1, 4);

                case 4:
                  hexlength = _context.sent;
                  _context.next = 10;
                  break;

                case 7:
                  _context.prev = 7;
                  _context.t0 = _context['catch'](1);
                  return _context.abrupt('return', null);

                case 10:
                  length = parseInt(hexlength.toString('utf8'), 16);
                  // skip over flush packets

                  if (!(length === 0)) {
                    _context.next = 13;
                    break;
                  }

                  return _context.abrupt('return', read());

                case 13:
                  _context.next = 15;
                  return gartal.readBytes(stream$$1, length - 4);

                case 15:
                  bytes = _context.sent;
                  return _context.abrupt('return', bytes);

                case 17:
                case 'end':
                  return _context.stop();
              }
            }
          }, _callee, this, [[1, 7]]);
        }));

        function read() {
          return _ref.apply(this, arguments);
        }

        return read;
      }();
    }
  }]);

  return GitPktLine;
}();

function parseIDX(buffer$$1) {
  var reader = new BufferCursor(buffer$$1);
  var magic = reader.slice(4).toString('hex');
  // Check for IDX v2 magic number
  if (magic !== 'ff744f63') {
    return; // undefined
  }
  var version = reader.readUInt32BE();
  if (version !== 2) {
    throw new Error('Unable to read version ' + version + ' packfile IDX. (Only version 2 supported)');
  }
  // Verify checksums
  var shaComputed = shasum(buffer$$1.slice(0, -20));
  var shaClaimed = buffer$$1.slice(-20).toString('hex');
  if (shaClaimed !== shaComputed) {
    throw new Error('Invalid checksum in IDX buffer: expected ' + shaClaimed + ' but saw ' + shaComputed);
  }
  if (buffer$$1.byteLength > 2048 * 1024 * 1024) {
    throw new Error('To keep implementation simple, I haven\'t implemented the layer 5 feature needed to support packfiles > 2GB in size.');
  }
  var fanout = [];
  for (var i = 0; i < 256; i++) {
    fanout.push(reader.readUInt32BE());
  }
  var size = fanout[255];
  // For now we'll parse the whole thing. We can optimize later if we need to.
  var hashes = [];
  for (var _i = 0; _i < size; _i++) {
    hashes.push(reader.slice(20).toString('hex'));
  }
  var crcs = new _Map();
  for (var _i2 = 0; _i2 < size; _i2++) {
    crcs.set(hashes[_i2], reader.readUInt32BE());
  }
  var offsets = new _Map();
  for (var _i3 = 0; _i3 < size; _i3++) {
    offsets.set(hashes[_i3], reader.readUInt32BE());
  }
  var packfileSha = reader.slice(20).toString('hex');
  // This part is gratuitous, but since we lack good unzipping arbitrary streams with extra at the end in the browser...
  var lengths = _Array$from(offsets);
  lengths.sort(function (a, b) {
    return a[1] - b[1];
  }); // List objects in order by offset
  var sizes = new _Map();
  var slices = new _Map();
  for (var _i4 = 0; _i4 < size - 1; _i4++) {
    sizes.set(lengths[_i4][0], lengths[_i4 + 1][1] - lengths[_i4][1]);
    slices.set(lengths[_i4][0], [lengths[_i4][1], lengths[_i4 + 1][1]]);
  }
  slices.set(lengths[size - 1][0], [lengths[size - 1][1], undefined]);
  return { size: size, fanout: fanout, hashes: hashes, crcs: crcs, packfileSha: packfileSha, slices: slices };
}

var GitPackfile = function () {
  function GitPackfile(_ref) {
    var size = _ref.size,
        fanout = _ref.fanout,
        hashes = _ref.hashes,
        crcs = _ref.crcs,
        packfileSha = _ref.packfileSha,
        slices = _ref.slices,
        pack = _ref.pack;

    _classCallCheck(this, GitPackfile);

    // Compare checksums
    var shaClaimed = pack.slice(-20).toString('hex');
    if (packfileSha !== shaClaimed) {
      throw new Error('Invalid packfile shasum in IDX buffer: expected ' + packfileSha + ' but saw ' + shaClaimed);
    }
    _Object$assign(this, {
      size: size,
      fanout: fanout,
      hashes: hashes,
      crcs: crcs,
      packfileSha: packfileSha,
      slices: slices,
      pack: pack
    });
  }

  _createClass(GitPackfile, [{
    key: 'read',
    value: function () {
      var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref2 /*: {oid: string} */) {
        var _pack;

        var oid = _ref2.oid;
        var raw;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (this.slices.has(oid)) {
                  _context.next = 2;
                  break;
                }

                return _context.abrupt('return', null);

              case 2:
                raw = (_pack = this.pack).slice.apply(_pack, _toConsumableArray(this.slices.get(oid)));
                /*
                - The header is followed by number of object entries, each of
                 which looks like this:
                  (undeltified representation)
                 n-byte type and length (3-bit type, (n-1)*7+4-bit length)
                 compressed data
                  (deltified representation)
                 n-byte type and length (3-bit type, (n-1)*7+4-bit length)
                 20-byte base object name if OBJ_REF_DELTA or a negative relative
                offset from the delta object's position in the pack if this
                is an OBJ_OFS_DELTA object
                 compressed delta data
                  Observation: length of each object is encoded in a variable
                 length format and is not constrained to 32-bit or anything.
                */

              case 3:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function read(_x) {
        return _ref3.apply(this, arguments);
      }

      return read;
    }()
  }], [{
    key: 'fromIDX',
    value: function () {
      var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref4) {
        var idx = _ref4.idx,
            pack = _ref4.pack;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                return _context2.abrupt('return', new GitPackfile(_extends({ pack: pack }, parseIDX(idx))));

              case 1:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function fromIDX(_x2) {
        return _ref5.apply(this, arguments);
      }

      return fromIDX;
    }()
  }]);

  return GitPackfile;
}();

// @flow
/*::
import type {Stats} from 'fs'

type CacheEntryFlags = {
  assumeValid: boolean,
  extended: boolean,
  stage: number,
  nameLength: number
}

type CacheEntry = {
  ctime: Date,
  ctimeNanoseconds?: number,
  mtime: Date,
  mtimeNanoseconds?: number,
  dev: number,
  ino: number,
  mode: number,
  uid: number,
  gid: number,
  size: number,
  oid: string,
  flags: CacheEntryFlags,
  path: string
}
*/

// Extract 1-bit assume-valid, 1-bit extended flag, 2-bit merge state flag, 12-bit path length flag
function parseCacheEntryFlags(bits /*: number */) /*: CacheEntryFlags */{
  return {
    assumeValid: Boolean(bits & 32768),
    extended: Boolean(bits & 16384),
    stage: (bits & 12288) >> 12,
    nameLength: bits & 4095
  };
}

function renderCacheEntryFlags(flags /*: CacheEntryFlags */) /*: number */{
  return (flags.assumeValid ? 32768 : 0) + (flags.extended ? 16384 : 0) + ((flags.stage & 3) << 12) + (flags.nameLength & 4095);
}

function parseBuffer(buffer$$1) {
  // Verify shasum
  var shaComputed = shasum(buffer$$1.slice(0, -20));
  var shaClaimed = buffer$$1.slice(-20).toString('hex');
  if (shaClaimed !== shaComputed) {
    throw new Error('Invalid checksum in GitIndex buffer: expected ' + shaClaimed + ' but saw ' + shaComputed);
  }
  var reader = new BufferCursor(buffer$$1);
  var _entries /*: Map<string, CacheEntry> */ = new _Map();
  var magic = reader.toString('utf8', 4);
  if (magic !== 'DIRC') {
    throw new Error('Inavlid dircache magic file number: ' + magic);
  }
  var version = reader.readUInt32BE();
  if (version !== 2) throw new Error('Unsupported dircache version: ' + version);
  var numEntries = reader.readUInt32BE();
  var i = 0;
  while (!reader.eof() && i < numEntries) {
    var entry = {};
    var ctimeSeconds = reader.readUInt32BE();
    var ctimeNanoseconds = reader.readUInt32BE();
    entry.ctime = new Date(ctimeSeconds * 1000 + ctimeNanoseconds / 1000000);
    entry.ctimeNanoseconds = ctimeNanoseconds;
    var mtimeSeconds = reader.readUInt32BE();
    var mtimeNanoseconds = reader.readUInt32BE();
    entry.mtime = new Date(mtimeSeconds * 1000 + mtimeNanoseconds / 1000000);
    entry.mtimeNanoseconds = mtimeNanoseconds;
    entry.dev = reader.readUInt32BE();
    entry.ino = reader.readUInt32BE();
    entry.mode = reader.readUInt32BE();
    entry.uid = reader.readUInt32BE();
    entry.gid = reader.readUInt32BE();
    entry.size = reader.readUInt32BE();
    entry.oid = reader.slice(20).toString('hex');
    var flags = reader.readUInt16BE();
    entry.flags = parseCacheEntryFlags(flags);
    // TODO: handle if (version === 3 && entry.flags.extended)
    var pathlength = buffer$$1.indexOf(0, reader.tell() + 1) - reader.tell();
    if (pathlength < 1) throw new Error('Got a path length of: ' + pathlength);
    entry.path = reader.toString('utf8', pathlength);
    // The next bit is awkward. We expect 1 to 8 null characters
    var tmp = reader.readUInt8();
    if (tmp !== 0) {
      throw new Error('Expected 1-8 null characters but got \'' + tmp + '\'');
    }
    var numnull = 1;
    while (!reader.eof() && reader.readUInt8() === 0 && numnull < 9) {
      numnull++;
    }reader.seek(reader.tell() - 1);
    // end of awkward part
    _entries.set(entry.path, entry);
    i++;
  }
  return _entries;
}

var GitIndex = function () {
  /*::
   _entries: Map<string, CacheEntry>
   _dirty: boolean // Used to determine if index needs to be saved to filesystem
   */
  function GitIndex(index /*: any */) {
    _classCallCheck(this, GitIndex);

    this._dirty = false;
    if (buffer.Buffer.isBuffer(index)) {
      this._entries = parseBuffer(index);
    } else if (index === null) {
      this._entries = new _Map();
    } else {
      throw new Error('invalid type passed to GitIndex constructor');
    }
  }

  _createClass(GitIndex, [{
    key: _Symbol$iterator,
    value: _regeneratorRuntime.mark(function value() {
      var _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry;

      return _regeneratorRuntime.wrap(function value$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _iteratorNormalCompletion = true;
              _didIteratorError = false;
              _iteratorError = undefined;
              _context.prev = 3;
              _iterator = _getIterator(this.entries);

            case 5:
              if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                _context.next = 12;
                break;
              }

              entry = _step.value;
              _context.next = 9;
              return entry;

            case 9:
              _iteratorNormalCompletion = true;
              _context.next = 5;
              break;

            case 12:
              _context.next = 18;
              break;

            case 14:
              _context.prev = 14;
              _context.t0 = _context['catch'](3);
              _didIteratorError = true;
              _iteratorError = _context.t0;

            case 18:
              _context.prev = 18;
              _context.prev = 19;

              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }

            case 21:
              _context.prev = 21;

              if (!_didIteratorError) {
                _context.next = 24;
                break;
              }

              throw _iteratorError;

            case 24:
              return _context.finish(21);

            case 25:
              return _context.finish(18);

            case 26:
            case 'end':
              return _context.stop();
          }
        }
      }, value, this, [[3, 14, 18, 26], [19,, 21, 25]]);
    })
  }, {
    key: 'insert',
    value: function insert(_ref) /*: {filepath: string, stats: Stats, oid: string } */{
      var filepath = _ref.filepath,
          stats = _ref.stats,
          oid = _ref.oid;

      var entry = {
        ctime: stats.ctime,
        mtime: stats.mtime,
        dev: stats.dev,
        ino: stats.ino,
        mode: stats.mode,
        uid: stats.uid,
        gid: stats.gid,
        size: stats.size,
        path: filepath,
        oid: oid,
        flags: {
          assumeValid: false,
          extended: false,
          stage: 0,
          nameLength: filepath.length < 0xfff ? filepath.length : 0xfff
        }
      };
      this._entries.set(entry.path, entry);
      this._dirty = true;
    }
  }, {
    key: 'delete',
    value: function _delete(_ref2 /*: {filepath: string} */) {
      var filepath = _ref2.filepath;

      if (this._entries.has(filepath)) {
        this._entries.delete(filepath);
      } else {
        var _iteratorNormalCompletion2 = true;
        var _didIteratorError2 = false;
        var _iteratorError2 = undefined;

        try {
          for (var _iterator2 = _getIterator(this._entries.keys()), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
            var key$$1 = _step2.value;

            if (key$$1.startsWith(filepath + '/')) {
              this._entries.delete(key$$1);
            }
          }
        } catch (err) {
          _didIteratorError2 = true;
          _iteratorError2 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }
          } finally {
            if (_didIteratorError2) {
              throw _iteratorError2;
            }
          }
        }
      }
      this._dirty = true;
    }
  }, {
    key: 'clear',
    value: function clear() {
      this._entries.clear();
      this._dirty = true;
    }
  }, {
    key: 'render',
    value: function render() {
      return this.entries.map(function (entry) {
        return entry.mode.toString(8) + ' ' + entry.oid + '    ' + entry.path;
      }).join('\n');
    }
  }, {
    key: 'toObject',
    value: function toObject() {
      var header = buffer.Buffer.alloc(12);
      var writer = new BufferCursor(header);
      writer.write('DIRC', 4, 'utf8');
      writer.writeUInt32BE(2);
      writer.writeUInt32BE(this.entries.length);
      var body = buffer.Buffer.concat(this.entries.map(function (entry) {
        // the fixed length + the filename + at least one null char => align by 8
        var length = Math.ceil((62 + entry.path.length + 1) / 8) * 8;
        var written = buffer.Buffer.alloc(length);
        var writer = new BufferCursor(written);
        var ctimeMilliseconds = entry.ctime.valueOf();
        var ctimeSeconds = Math.floor(ctimeMilliseconds / 1000);
        var ctimeNanoseconds = entry.ctimeNanoseconds || ctimeMilliseconds * 1000000 - ctimeSeconds * 1000000 * 1000;
        var mtimeMilliseconds = entry.mtime.valueOf();
        var mtimeSeconds = Math.floor(mtimeMilliseconds / 1000);
        var mtimeNanoseconds = entry.mtimeNanoseconds || mtimeMilliseconds * 1000000 - mtimeSeconds * 1000000 * 1000;
        writer.writeUInt32BE(ctimeSeconds);
        writer.writeUInt32BE(ctimeNanoseconds);
        writer.writeUInt32BE(mtimeSeconds);
        writer.writeUInt32BE(mtimeNanoseconds);
        writer.writeUInt32BE(entry.dev);
        writer.writeUInt32BE(entry.ino);
        writer.writeUInt32BE(entry.mode);
        writer.writeUInt32BE(entry.uid);
        writer.writeUInt32BE(entry.gid);
        writer.writeUInt32BE(entry.size);
        writer.write(entry.oid, 20, 'hex');
        writer.writeUInt16BE(renderCacheEntryFlags(entry.flags));
        writer.write(entry.path, entry.path.length, 'utf8');
        return written;
      }));
      var main = buffer.Buffer.concat([header, body]);
      var sum = shasum(main);
      return buffer.Buffer.concat([main, buffer.Buffer.from(sum, 'hex')]);
    }
  }, {
    key: 'entries',
    get: function get() /*: Array<CacheEntry> */{
      return sortby([].concat(_toConsumableArray(this._entries.values())), 'path');
    }
  }], [{
    key: 'from',
    value: function from(buffer$$1) {
      return new GitIndex(buffer$$1);
    }
  }]);

  return GitIndex;
}();

// @flow
/*::
type TreeEntry = {
  mode: string,
  path: string,
  oid: string,
  type?: string
}
*/

function parseBuffer$1(buffer$$1) /*: Array<TreeEntry> */{
  var _entries = [];
  var cursor = 0;
  while (cursor < buffer$$1.length) {
    var space = buffer$$1.indexOf(32, cursor);
    if (space === -1) {
      throw new Error('GitTree: Error parsing buffer at byte location ' + cursor + ': Could not find the next space character.');
    }
    var nullchar = buffer$$1.indexOf(0, cursor);
    if (nullchar === -1) {
      throw new Error('GitTree: Error parsing buffer at byte location ' + cursor + ': Could not find the next null character.');
    }
    var mode = buffer$$1.slice(cursor, space).toString('utf8');
    if (mode === '40000') mode = '040000'; // makes it line up neater in printed output
    var type = mode === '040000' ? 'tree' : 'blob';
    var path$$1 = buffer$$1.slice(space + 1, nullchar).toString('utf8');
    var oid = buffer$$1.slice(nullchar + 1, nullchar + 21).toString('hex');
    cursor = nullchar + 21;
    _entries.push({ mode: mode, path: path$$1, oid: oid, type: type });
  }
  return _entries;
}

function nudgeIntoShape(entry) {
  if (!entry.oid && entry.sha) {
    entry.oid = entry.sha; // Github
  }
  if (typeof entry.mode === 'number') {
    entry.mode = entry.mode.toString(8); // index
  }
  if (!entry.type) {
    entry.type = 'blob'; // index
  }
  return entry;
}

var GitTree = function () {
  /*::
  _entries: Array<TreeEntry>
  */
  function GitTree(entries /*: any */) {
    _classCallCheck(this, GitTree);

    if (buffer.Buffer.isBuffer(entries)) {
      this._entries = parseBuffer$1(entries);
    } else if (Array.isArray(entries)) {
      this._entries = entries.map(nudgeIntoShape);
    } else {
      throw new Error('invalid type passed to GitTree constructor');
    }
  }

  _createClass(GitTree, [{
    key: 'render',
    value: function render() {
      return this._entries.map(function (entry) {
        return entry.mode + ' ' + entry.type + ' ' + entry.oid + '    ' + entry.path;
      }).join('\n');
    }
  }, {
    key: 'toObject',
    value: function toObject() {
      return buffer.Buffer.concat(this._entries.map(function (entry) {
        var mode = buffer.Buffer.from(entry.mode.replace(/^0/, ''));
        var space = buffer.Buffer.from(' ');
        var path$$1 = buffer.Buffer.from(entry.path, { encoding: 'utf8' });
        var nullchar = buffer.Buffer.from([0]);
        var oid = buffer.Buffer.from(entry.oid.match(/../g).map(function (n) {
          return parseInt(n, 16);
        }));
        return buffer.Buffer.concat([mode, space, path$$1, nullchar, oid]);
      }));
    }
  }, {
    key: 'entries',
    value: function entries() {
      return this._entries;
    }
  }, {
    key: _Symbol$iterator,
    value: _regeneratorRuntime.mark(function value() {
      var _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry;

      return _regeneratorRuntime.wrap(function value$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              _iteratorNormalCompletion = true;
              _didIteratorError = false;
              _iteratorError = undefined;
              _context.prev = 3;
              _iterator = _getIterator(this._entries);

            case 5:
              if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                _context.next = 12;
                break;
              }

              entry = _step.value;
              _context.next = 9;
              return entry;

            case 9:
              _iteratorNormalCompletion = true;
              _context.next = 5;
              break;

            case 12:
              _context.next = 18;
              break;

            case 14:
              _context.prev = 14;
              _context.t0 = _context['catch'](3);
              _didIteratorError = true;
              _iteratorError = _context.t0;

            case 18:
              _context.prev = 18;
              _context.prev = 19;

              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }

            case 21:
              _context.prev = 21;

              if (!_didIteratorError) {
                _context.next = 24;
                break;
              }

              throw _iteratorError;

            case 24:
              return _context.finish(21);

            case 25:
              return _context.finish(18);

            case 26:
            case 'end':
              return _context.stop();
          }
        }
      }, value, this, [[3, 14, 18, 26], [19,, 21, 25]]);
    })
  }], [{
    key: 'from',
    value: function from(tree) {
      return new GitTree(tree);
    }
  }]);

  return GitTree;
}();



var models = Object.freeze({
	GitCommit: GitCommit,
	GitConfig: GitConfig,
	GitObject: GitObject,
	GitPktLine: GitPktLine,
	GitPackfile: GitPackfile,
	GitIndex: GitIndex,
	GitTree: GitTree
});

var fs;
var defaultfs = function () {
  return fs;
};
function setfs(newFs) {
  fs = newFs;
}

var rm = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(filepath) {
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return pify(defaultfs().unlink)(filepath);

          case 3:
            _context.next = 9;
            break;

          case 5:
            _context.prev = 5;
            _context.t0 = _context['catch'](0);

            if (!(_context.t0.code !== 'ENOENT')) {
              _context.next = 9;
              break;
            }

            throw _context.t0;

          case 9:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[0, 5]]);
  }));

  return function rm(_x) {
    return _ref.apply(this, arguments);
  };
}();

// An async exists variant
var exists = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(file, options) {
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            return _context.abrupt('return', new _Promise(function (resolve, reject) {
              defaultfs().stat(file, function (err, stats) {
                if (err) return err.code === 'ENOENT' ? resolve(false) : reject(err);
                resolve(true);
              });
            }));

          case 1:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function exists(_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();

// @flow
/*::
type Node = {
  type: string,
  fullpath: string,
  basename: string,
  metadata: Object, // mode, oid
  parent?: Node,
  children: Array<Node>
}
*/

function flatFileListToDirectoryStructure(files /*: Array<{path: string}> */
) /*: Node|void */{
  var inodes /*: Map<string, Node> */ = new _Map();
  var mkdir = function mkdir(name) /*: Node|void */{
    if (!inodes.has(name)) {
      var dir /*: Node */ = {
        type: 'tree',
        fullpath: name,
        basename: path.basename(name),
        metadata: {},
        children: []
      };
      inodes.set(name, dir);
      // This recursively generates any missing parent folders.
      // We do it after we've added the inode to the set so that
      // we don't recurse infinitely trying to create the root '.' dirname.
      dir.parent = mkdir(path.dirname(name));
      if (dir.parent && dir.parent !== dir) dir.parent.children.push(dir);
    }
    return inodes.get(name);
  };

  var mkfile = function mkfile(name, metadata) /*: Node|void */{
    if (!inodes.has(name)) {
      var file /*: Node */ = {
        type: 'blob',
        fullpath: name,
        basename: path.basename(name),
        metadata: metadata,
        // This recursively generates any missing parent folders.
        parent: mkdir(path.dirname(name)),
        children: []
      };
      if (file.parent) file.parent.children.push(file);
      inodes.set(name, file);
    }
    return inodes.get(name);
  };

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = _getIterator(files), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var file = _step.value;

      mkfile(file.path, file);
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return inodes.get('.');
}

// Based on an elegant concurrent recursive solution from SO
// https://stackoverflow.com/a/45130990/2168416
var readDirAsFlatFileList = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(dir) {
    var _this = this;

    var subdirs, files;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return pify(defaultfs().readdir)(dir);

          case 2:
            subdirs = _context2.sent;
            _context2.next = 5;
            return _Promise.all(subdirs.map(function () {
              var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(subdir) {
                var res;
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        res = dir + '/' + subdir;
                        _context.next = 3;
                        return pify(defaultfs().stat)(res);

                      case 3:
                        if (!_context.sent.isDirectory()) {
                          _context.next = 7;
                          break;
                        }

                        _context.t0 = readDirAsFlatFileList(res);
                        _context.next = 8;
                        break;

                      case 7:
                        _context.t0 = res;

                      case 8:
                        return _context.abrupt('return', _context.t0);

                      case 9:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, _this);
              }));

              return function (_x2) {
                return _ref2.apply(this, arguments);
              };
            }()));

          case 5:
            files = _context2.sent;
            return _context2.abrupt('return', files.reduce(function (a, f) {
              return a.concat(f);
            }, []));

          case 7:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function readDirAsFlatFileList(_x) {
    return _ref.apply(this, arguments);
  };
}();

var sleep = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(ms) {
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            return _context.abrupt("return", new _Promise(function (resolve, reject) {
              return setTimeout(resolve, ms);
            }));

          case 1:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function sleep(_x) {
    return _ref.apply(this, arguments);
  };
}();

// @flow
// This is modeled after the lockfile strategy used by the git source code.
var delayedReleases = new _Map();

var lock = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(filename /*: string */
  ) {
    var triesLeft /*: number */ = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 3;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (!delayedReleases.has(filename)) {
              _context.next = 4;
              break;
            }

            clearTimeout(delayedReleases.get(filename));
            delayedReleases.delete(filename);
            return _context.abrupt('return');

          case 4:
            if (!(triesLeft === 0)) {
              _context.next = 6;
              break;
            }

            throw new Error('Unable to acquire lockfile \'' + filename + '\'. Exhausted tries.');

          case 6:
            _context.prev = 6;
            _context.next = 9;
            return pify(defaultfs().mkdir)(filename + '.lock');

          case 9:
            _context.next = 18;
            break;

          case 11:
            _context.prev = 11;
            _context.t0 = _context['catch'](6);

            if (!(_context.t0.code === 'EEXIST')) {
              _context.next = 18;
              break;
            }

            _context.next = 16;
            return sleep(100);

          case 16:
            _context.next = 18;
            return lock(filename, triesLeft - 1);

          case 18:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[6, 11]]);
  }));

  return function lock(_x2) {
    return _ref.apply(this, arguments);
  };
}();

var unlock = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(filename /*: string */
  ) {
    var _this = this;

    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            if (!delayedReleases.has(filename)) {
              _context3.next = 2;
              break;
            }

            throw new Error('Cannot double-release lockfile');

          case 2:
            // Basically, we lie and say it was deleted ASAP.
            // But really we wait a bit to see if you want to acquire it again.
            delayedReleases.set(filename, setTimeout(_asyncToGenerator(_regeneratorRuntime.mark(function _callee2() {
              return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                while (1) {
                  switch (_context2.prev = _context2.next) {
                    case 0:
                      delayedReleases.delete(filename);
                      _context2.next = 3;
                      return pify(defaultfs().rmdir)(filename + '.lock');

                    case 3:
                    case 'end':
                      return _context2.stop();
                  }
                }
              }, _callee2, _this);
            }))));

          case 3:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function unlock(_x4) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var mkdir = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(dirpath /*: string */) {
    var parent;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return pify(defaultfs().mkdir)(dirpath);

          case 3:
            return _context.abrupt('return');

          case 6:
            _context.prev = 6;
            _context.t0 = _context['catch'](0);

            if (!(_context.t0 === null)) {
              _context.next = 10;
              break;
            }

            return _context.abrupt('return');

          case 10:
            if (!(_context.t0.code === 'EEXIST')) {
              _context.next = 12;
              break;
            }

            return _context.abrupt('return');

          case 12:
            if (!(_context.t0.code === 'ENOENT')) {
              _context.next = 20;
              break;
            }

            parent = path.dirname(dirpath);
            // Check to see if we've gone too far

            if (!(parent === '.' || parent === '/' || parent === dirpath)) {
              _context.next = 16;
              break;
            }

            throw _context.t0;

          case 16:
            _context.next = 18;
            return mkdir(parent);

          case 18:
            _context.next = 20;
            return mkdir(dirpath);

          case 20:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[0, 6]]);
  }));

  return function mkdir(_x) {
    return _ref.apply(this, arguments);
  };
}();

var mkdirs = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(dirlist /*: string[] */) {
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            return _context2.abrupt('return', _Promise.all(dirlist.map(mkdir)));

          case 1:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function mkdirs(_x2) {
    return _ref2.apply(this, arguments);
  };
}();

// An async readFile variant that returns null instead of throwing errors
var read = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(file, options) {
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            return _context.abrupt('return', new _Promise(function (resolve, reject) {
              defaultfs().readFile(file, options, function (err, file) {
                return err ? resolve(null) : resolve(file);
              });
            }));

          case 1:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function read(_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();

// @flow
// An async writeFile variant that automatically creates missing directories,
// and returns null instead of throwing errors.
var write = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(filepath /*: string */
  , contents /*: string|Buffer */
  ) {
    var options /*: Object */ = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return pify(defaultfs().writeFile)(filepath, contents, options);

          case 3:
            return _context.abrupt('return');

          case 6:
            _context.prev = 6;
            _context.t0 = _context['catch'](0);
            _context.next = 10;
            return mkdir(path.dirname(filepath));

          case 10:
            _context.next = 12;
            return pify(defaultfs().writeFile)(filepath, contents, options);

          case 12:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[0, 6]]);
  }));

  return function write(_x2, _x3) {
    return _ref.apply(this, arguments);
  };
}();

var name = "isomorphic-git";
var version = "0.0.27";
var description = "A pure JavaScript implementation of git for node and browsers!";
var main = "dist/for-node/";
var browser = "dist/for-browserify/";
var module$1 = "dist/for-future/";
var unpkg = "dist/bundle.umd.min.js";
var bin = { "isogit": "./cli.js" };
var engines = { "node": ">=7.6.0" };
var scripts = { "start": "nps", "test": "nps test", "precommit": "nps format toc" };
var repository = { "type": "git", "url": "git+https://github.com/wmhilton/isomorphic-git.git" };
var keywords = ["git"];
var author = "William Hilton <wmhilton@gmail.com>";
var license = "Unlicense";
var bugs = { "url": "https://github.com/wmhilton/isomorphic-git/issues" };
var homepage = "https://github.com/wmhilton/isomorphic-git#readme";
var files = ["dist", "cli.js"];
var dependencies = { "async-lock": "^1.0.0", "await-stream-ready": "^1.0.1", "babel-runtime": "^6.26.0", "buffer": "^5.0.7", "buffer-peek-stream": "^1.0.1", "buffercursor": "0.0.12", "gartal": "^1.1.2", "git-apply-delta": "0.0.7", "git-list-pack": "0.0.10", "ignore": "^3.3.6", "ini": "^1.3.4", "marky": "^1.2.0", "minimisted": "^2.0.0", "openpgp": "^2.5.10", "pad": "^2.0.1", "pako": "^1.0.5", "pify": "^3.0.0", "shasum": "^1.0.2", "simple-concat": "^1.0.0", "simple-get": "^2.7.0", "thru": "git+https://github.com/wmhilton-contrib/thru.git#master" };
var devDependencies = { "babel-plugin-external-helpers": "^6.22.0", "babel-plugin-transform-es2015-modules-commonjs": "^6.24.1", "babel-plugin-transform-object-rest-spread": "^6.23.0", "babel-plugin-transform-runtime": "^6.23.0", "babel-preset-env": "^1.6.0", "babel-preset-flow": "^6.23.0", "ban-sensitive-files": "^1.9.0", "browserfs": "^1.4.3", "browserify": "^14.4.0", "browserify-shim": "^3.8.14", "codecov": "^3.0.0", "doctoc": "^1.3.0", "husky": "^0.14.3", "jest": "^21.2.1", "jest-fixtures": "^0.6.0", "jsonfile": "^4.0.0", "karma": "^1.7.1", "karma-browserify": "^5.1.1", "karma-chrome-launcher": "^2.2.0", "karma-firefox-launcher": "^1.0.1", "karma-sauce-launcher": "^1.2.0", "karma-tap": "^3.1.1", "lodash": "^4.17.4", "nock": "^9.0.17", "npm-run-all": "^4.1.1", "nps": "^5.7.1", "nps-utils": "^1.4.0", "parse-header-stream": "^1.1.1", "prettier-standard": "^7.0.1", "rollup": "^0.51.6", "rollup-plugin-babel": "^3.0.2", "rollup-plugin-json": "^2.3.0", "standard": "^10.0.3", "stream-equal": "^1.0.1", "tape": "^4.8.0", "uglify-es": "^3.1.2", "watchify": "^3.9.0" };
var ava = { "source": ["dist/for-node/*"] };
var browserify = { "transform": ["browserify-shim"] };
var testling = { "files": "testling/basic-test.js", "browsers": ["chrome/latest", "firefox/latest", "ie/latest"] };
var jest = { "testPathIgnorePatterns": ["__helpers__"], "testEnvironment": "node" };
var _package = {
	name: name,
	version: version,
	description: description,
	main: main,
	browser: browser,
	module: module$1,
	unpkg: unpkg,
	bin: bin,
	engines: engines,
	scripts: scripts,
	repository: repository,
	keywords: keywords,
	author: author,
	license: license,
	bugs: bugs,
	homepage: homepage,
	files: files,
	dependencies: dependencies,
	devDependencies: devDependencies,
	ava: ava,
	browserify: browserify,
	testling: testling,
	jest: jest,
	"browserify-shim": { "fs": "global:fs" }
};

var _package$1 = Object.freeze({
	name: name,
	version: version,
	description: description,
	main: main,
	browser: browser,
	module: module$1,
	unpkg: unpkg,
	bin: bin,
	engines: engines,
	scripts: scripts,
	repository: repository,
	keywords: keywords,
	author: author,
	license: license,
	bugs: bugs,
	homepage: homepage,
	files: files,
	dependencies: dependencies,
	devDependencies: devDependencies,
	ava: ava,
	browserify: browserify,
	testling: testling,
	jest: jest,
	default: _package
});



var utils = Object.freeze({
	rm: rm,
	exists: exists,
	flatFileListToDirectoryStructure: flatFileListToDirectoryStructure,
	readDirAsFlatFileList: readDirAsFlatFileList,
	fs: defaultfs,
	setfs: setfs,
	lock: lock,
	unlock: unlock,
	mkdirs: mkdirs,
	read: read,
	sleep: sleep,
	write: write,
	pkg: _package$1
});

// @flow
var GitConfigManager = function () {
  function GitConfigManager() {
    _classCallCheck(this, GitConfigManager);
  }

  _createClass(GitConfigManager, null, [{
    key: 'get',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
        var gitdir = _ref.gitdir;
        var text;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return read(gitdir + '/config', { encoding: 'utf8' });

              case 2:
                text = _context.sent;
                return _context.abrupt('return', GitConfig.from(text));

              case 4:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function get(_x) {
        return _ref2.apply(this, arguments);
      }

      return get;
    }()
  }, {
    key: 'save',
    value: function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref3) {
        var gitdir = _ref3.gitdir,
            config = _ref3.config;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return write(gitdir + '/config', config.toString(), {
                  encoding: 'utf8'
                });

              case 2:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function save(_x2) {
        return _ref4.apply(this, arguments);
      }

      return save;
    }()
  }]);

  return GitConfigManager;
}();

// @flow
// TODO: Add file locks.
var GitShallowManager = function () {
  function GitShallowManager() {
    _classCallCheck(this, GitShallowManager);
  }

  _createClass(GitShallowManager, null, [{
    key: 'read',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
        var gitdir = _ref.gitdir;
        var oids, text;
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                oids = new _Set();
                _context.next = 3;
                return read(path.join(gitdir, 'shallow'), { encoding: 'utf8' });

              case 3:
                text = _context.sent;

                if (!(text === null)) {
                  _context.next = 6;
                  break;
                }

                return _context.abrupt('return', oids);

              case 6:
                text.trim().split('\n').map(function (oid) {
                  return oids.add(oid);
                });
                return _context.abrupt('return', oids);

              case 8:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function read$$1(_x) {
        return _ref2.apply(this, arguments);
      }

      return read$$1;
    }()
  }, {
    key: 'write',
    value: function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref3) {
        var gitdir = _ref3.gitdir,
            oids = _ref3.oids;

        var text, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, oid;

        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                text = '';
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context2.prev = 4;

                for (_iterator = _getIterator(oids); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                  oid = _step.value;

                  text += oid + '\n';
                }
                _context2.next = 12;
                break;

              case 8:
                _context2.prev = 8;
                _context2.t0 = _context2['catch'](4);
                _didIteratorError = true;
                _iteratorError = _context2.t0;

              case 12:
                _context2.prev = 12;
                _context2.prev = 13;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 15:
                _context2.prev = 15;

                if (!_didIteratorError) {
                  _context2.next = 18;
                  break;
                }

                throw _iteratorError;

              case 18:
                return _context2.finish(15);

              case 19:
                return _context2.finish(12);

              case 20:
                _context2.next = 22;
                return write(path.join(gitdir, 'shallow'), text, {
                  encoding: 'utf8'
                });

              case 22:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this, [[4, 8, 12, 20], [13,, 15, 19]]);
      }));

      function write$$1(_x2) {
        return _ref4.apply(this, arguments);
      }

      return write$$1;
    }()
  }]);

  return GitShallowManager;
}();

// @flow
// import LockManager from 'travix-lock-manager'
// import Lock from '../utils'

// TODO: replace with an LRU cache?
var map /*: Map<string, GitIndex> */ = new _Map();
// const lm = new LockManager()
var lock$1 = new AsyncLock();

var GitIndexManager = function () {
  function GitIndexManager() {
    _classCallCheck(this, GitIndexManager);
  }

  _createClass(GitIndexManager, null, [{
    key: 'acquire',
    value: function () {
      var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(filepath, closure) {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return lock$1.acquire(filepath, _asyncToGenerator(_regeneratorRuntime.mark(function _callee() {
                  var index, rawIndexFile, buffer$$1;
                  return _regeneratorRuntime.wrap(function _callee$(_context) {
                    while (1) {
                      switch (_context.prev = _context.next) {
                        case 0:
                          index = map.get(filepath);

                          if (!(index === undefined)) {
                            _context.next = 7;
                            break;
                          }

                          _context.next = 4;
                          return read(filepath);

                        case 4:
                          rawIndexFile = _context.sent;

                          index = GitIndex.from(rawIndexFile);
                          // cache the GitIndex object so we don't need to re-read it
                          // every time.
                          // TODO: save the stat data for the index so we know whether
                          // the cached file is stale (modified by an outside process).
                          map.set(filepath, index);
                          // await fileLock.cancel()

                        case 7:
                          _context.next = 9;
                          return closure(index);

                        case 9:
                          if (!index._dirty) {
                            _context.next = 14;
                            break;
                          }

                          // Acquire a file lock while we're writing the index file
                          // let fileLock = await Lock(filepath)
                          buffer$$1 = index.toObject();
                          _context.next = 13;
                          return write(filepath, buffer$$1);

                        case 13:
                          index._dirty = false;

                        case 14:
                          // For now, discard our cached object so that external index
                          // manipulation is picked up. TODO: use lstat and compare
                          // file times to determine if our cached object should be
                          // discarded.
                          map.delete(filepath);

                        case 15:
                        case 'end':
                          return _context.stop();
                      }
                    }
                  }, _callee, this);
                })));

              case 2:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function acquire(_x, _x2) {
        return _ref.apply(this, arguments);
      }

      return acquire;
    }()
  }]);

  return GitIndexManager;
}();

// @flow
// I'm putting this in a Manager because I reckon it could benefit
// from a LOT of cacheing.

// TODO: Implement .git/info/exclude

var GitIgnoreManager = function () {
  function GitIgnoreManager() {
    _classCallCheck(this, GitIgnoreManager);
  }

  _createClass(GitIgnoreManager, null, [{
    key: 'isIgnored',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
        var gitdir = _ref.gitdir,
            workdir = _ref.workdir,
            pathname = _ref.pathname;

        var pairs, pieces, i, dir, file, ignoredStatus, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, p, _file, ign, unign, parentdir;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                pairs = [{
                  gitignore: path.join(workdir, '.gitignore'),
                  pathname: pathname
                }];
                pieces = pathname.split('/');

                for (i = 1; i < pieces.length; i++) {
                  dir = pieces.slice(0, i).join('/');
                  file = pieces.slice(i).join('/');

                  pairs.push({
                    gitignore: path.join(workdir, dir, '.gitignore'),
                    pathname: file
                  });
                }
                ignoredStatus = false;
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context.prev = 7;
                _iterator = _getIterator(pairs);

              case 9:
                if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                  _context.next = 31;
                  break;
                }

                p = _step.value;
                _file = void 0;
                _context.prev = 12;
                _context.next = 15;
                return read(p.gitignore, 'utf8');

              case 15:
                _file = _context.sent;
                _context.next = 22;
                break;

              case 18:
                _context.prev = 18;
                _context.t0 = _context['catch'](12);

                if (!(_context.t0.code === 'NOENT')) {
                  _context.next = 22;
                  break;
                }

                return _context.abrupt('continue', 28);

              case 22:
                ign = ignore().add(_file);
                unign = ignore().add('**\n' + _file);
                // If the parent directory is excluded, we are done.
                // "It is not possible to re-include a file if a parent directory of that file is excluded. Git doesn’t list excluded directories for performance reasons, so any patterns on contained files have no effect, no matter where they are defined."
                // source: https://git-scm.com/docs/gitignore

                parentdir = path.dirname(p.pathname);

                if (!ign.ignores(parentdir)) {
                  _context.next = 27;
                  break;
                }

                return _context.abrupt('return', true);

              case 27:
                // If the file is currently ignored, test for UNignoring.
                if (ignoredStatus) {
                  ignoredStatus = unign.ignores(p.pathname);
                } else {
                  ignoredStatus = ign.ignores(p.pathname);
                }

              case 28:
                _iteratorNormalCompletion = true;
                _context.next = 9;
                break;

              case 31:
                _context.next = 37;
                break;

              case 33:
                _context.prev = 33;
                _context.t1 = _context['catch'](7);
                _didIteratorError = true;
                _iteratorError = _context.t1;

              case 37:
                _context.prev = 37;
                _context.prev = 38;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 40:
                _context.prev = 40;

                if (!_didIteratorError) {
                  _context.next = 43;
                  break;
                }

                throw _iteratorError;

              case 43:
                return _context.finish(40);

              case 44:
                return _context.finish(37);

              case 45:
                return _context.abrupt('return', ignoredStatus);

              case 46:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this, [[7, 33, 37, 45], [12, 18], [38,, 40, 44]]);
      }));

      function isIgnored(_x) {
        return _ref2.apply(this, arguments);
      }

      return isIgnored;
    }()
  }]);

  return GitIgnoreManager;
}();

// @flow
var GitObjectManager = function () {
  function GitObjectManager() {
    _classCallCheck(this, GitObjectManager);
  }

  _createClass(GitObjectManager, null, [{
    key: 'read',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref /*: {gitdir: string, oid: string} */) {
        var gitdir = _ref.gitdir,
            oid = _ref.oid;

        var file, text, _GitObject$unwrap, type, object;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return read(gitdir + '/objects/' + oid.slice(0, 2) + '/' + oid.slice(2));

              case 2:
                file = _context.sent;

                if (file) {
                  _context.next = 12;
                  break;
                }

                _context.next = 6;
                return read(gitdir + '/shallow', { encoding: 'utf8' });

              case 6:
                text = _context.sent;

                if (!(text !== null && text.includes(oid))) {
                  _context.next = 11;
                  break;
                }

                throw new Error('Failed to read git object with oid ' + oid + ' because it is a shallow commit');

              case 11:
                throw new Error('Failed to read git object with oid ' + oid);

              case 12:
                _GitObject$unwrap = GitObject.unwrap({ oid: oid, file: file }), type = _GitObject$unwrap.type, object = _GitObject$unwrap.object;
                return _context.abrupt('return', { type: type, object: object });

              case 14:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function read$$1(_x) {
        return _ref2.apply(this, arguments);
      }

      return read$$1;
    }()
  }, {
    key: 'hash',
    value: function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref3) {
        var gitdir = _ref3.gitdir,
            type = _ref3.type,
            object = _ref3.object;
        var buffer$$1, oid;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                buffer$$1 = buffer.Buffer.concat([buffer.Buffer.from(type + ' '), buffer.Buffer.from(object.byteLength.toString()), buffer.Buffer.from([0]), buffer.Buffer.from(object)]);
                oid = shasum(buffer$$1);
                return _context2.abrupt('return', oid);

              case 3:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function hash(_x2) {
        return _ref4.apply(this, arguments);
      }

      return hash;
    }()
  }, {
    key: 'write',
    value: function () {
      var _ref6 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref5) {
        var gitdir = _ref5.gitdir,
            type = _ref5.type,
            object = _ref5.object;

        var _GitObject$wrap, file, oid, filepath;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _GitObject$wrap = GitObject.wrap({ type: type, object: object }), file = _GitObject$wrap.file, oid = _GitObject$wrap.oid;
                filepath = gitdir + '/objects/' + oid.slice(0, 2) + '/' + oid.slice(2);
                // Don't overwrite existing git objects - this helps avoid EPERM errors.
                // Although I don't know how we'd fix corrupted objects then. Perhaps delete them
                // on read?

                _context3.next = 4;
                return exists(filepath);

              case 4:
                if (_context3.sent) {
                  _context3.next = 7;
                  break;
                }

                _context3.next = 7;
                return write(filepath, file);

              case 7:
                return _context3.abrupt('return', oid);

              case 8:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function write$$1(_x3) {
        return _ref6.apply(this, arguments);
      }

      return write$$1;
    }()
  }]);

  return GitObjectManager;
}();

// @flow
// This is a convenience wrapper for reading and writing files in the 'refs' directory.
var GitRefManager = function () {
  function GitRefManager() {
    _classCallCheck(this, GitRefManager);
  }

  _createClass(GitRefManager, null, [{
    key: 'updateRemoteRefs',

    /* ::
    updateRemoteRefs : ({
      gitdir: string,
      remote: string,
      refs: Map<string, string>,
      symrefs: Map<string, string>
    }) => Promise<void>
    */
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
        var gitdir = _ref.gitdir,
            remote = _ref.remote,
            refs = _ref.refs,
            symrefs = _ref.symrefs;

        var _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, _ref3, _ref4, key$$1, value, actualRefsToWrite, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _ref5, _ref6, _key, _value, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3, _ref7, _ref8, _key2, _value2, branch, normalizeValue, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, _ref9, _ref10, _key3, _value3;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                // Validate input
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context.prev = 3;
                _iterator = _getIterator(refs);

              case 5:
                if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                  _context.next = 15;
                  break;
                }

                _ref3 = _step.value;
                _ref4 = _slicedToArray(_ref3, 2);
                key$$1 = _ref4[0];
                value = _ref4[1];

                if (value.match(/[0-9a-f]{40}/)) {
                  _context.next = 12;
                  break;
                }

                throw new Error('Unexpected ref contents: \'' + value + '\'');

              case 12:
                _iteratorNormalCompletion = true;
                _context.next = 5;
                break;

              case 15:
                _context.next = 21;
                break;

              case 17:
                _context.prev = 17;
                _context.t0 = _context['catch'](3);
                _didIteratorError = true;
                _iteratorError = _context.t0;

              case 21:
                _context.prev = 21;
                _context.prev = 22;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 24:
                _context.prev = 24;

                if (!_didIteratorError) {
                  _context.next = 27;
                  break;
                }

                throw _iteratorError;

              case 27:
                return _context.finish(24);

              case 28:
                return _context.finish(21);

              case 29:
                // Combine refs and symrefs giving symrefs priority
                actualRefsToWrite = new _Map();
                _iteratorNormalCompletion2 = true;
                _didIteratorError2 = false;
                _iteratorError2 = undefined;
                _context.prev = 33;

                for (_iterator2 = _getIterator(refs); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                  _ref5 = _step2.value;
                  _ref6 = _slicedToArray(_ref5, 2);
                  _key = _ref6[0];
                  _value = _ref6[1];
                  actualRefsToWrite.set(_key, _value);
                }_context.next = 41;
                break;

              case 37:
                _context.prev = 37;
                _context.t1 = _context['catch'](33);
                _didIteratorError2 = true;
                _iteratorError2 = _context.t1;

              case 41:
                _context.prev = 41;
                _context.prev = 42;

                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                  _iterator2.return();
                }

              case 44:
                _context.prev = 44;

                if (!_didIteratorError2) {
                  _context.next = 47;
                  break;
                }

                throw _iteratorError2;

              case 47:
                return _context.finish(44);

              case 48:
                return _context.finish(41);

              case 49:
                _iteratorNormalCompletion3 = true;
                _didIteratorError3 = false;
                _iteratorError3 = undefined;
                _context.prev = 52;
                for (_iterator3 = _getIterator(symrefs); !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                  _ref7 = _step3.value;
                  _ref8 = _slicedToArray(_ref7, 2);
                  _key2 = _ref8[0];
                  _value2 = _ref8[1];
                  branch = _value2.replace(/^refs\/heads\//, '');

                  actualRefsToWrite.set(_key2, 'ref: refs/remotes/' + remote + '/' + branch);
                }
                // Update files
                _context.next = 60;
                break;

              case 56:
                _context.prev = 56;
                _context.t2 = _context['catch'](52);
                _didIteratorError3 = true;
                _iteratorError3 = _context.t2;

              case 60:
                _context.prev = 60;
                _context.prev = 61;

                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                  _iterator3.return();
                }

              case 63:
                _context.prev = 63;

                if (!_didIteratorError3) {
                  _context.next = 66;
                  break;
                }

                throw _iteratorError3;

              case 66:
                return _context.finish(63);

              case 67:
                return _context.finish(60);

              case 68:
                normalizeValue = function normalizeValue(value) {
                  return value.trim() + '\n';
                };

                _iteratorNormalCompletion4 = true;
                _didIteratorError4 = false;
                _iteratorError4 = undefined;
                _context.prev = 72;
                _iterator4 = _getIterator(actualRefsToWrite);

              case 74:
                if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
                  _context.next = 86;
                  break;
                }

                _ref9 = _step4.value;
                _ref10 = _slicedToArray(_ref9, 2);
                _key3 = _ref10[0];
                _value3 = _ref10[1];

                // For some reason we trim these
                _key3 = _key3.replace(/^refs\/heads\//, '');
                _key3 = _key3.replace(/^refs\/tags\//, '');
                _context.next = 83;
                return write(path.join(gitdir, 'refs', 'remotes', remote, _key3), normalizeValue(_value3), 'utf8');

              case 83:
                _iteratorNormalCompletion4 = true;
                _context.next = 74;
                break;

              case 86:
                _context.next = 92;
                break;

              case 88:
                _context.prev = 88;
                _context.t3 = _context['catch'](72);
                _didIteratorError4 = true;
                _iteratorError4 = _context.t3;

              case 92:
                _context.prev = 92;
                _context.prev = 93;

                if (!_iteratorNormalCompletion4 && _iterator4.return) {
                  _iterator4.return();
                }

              case 95:
                _context.prev = 95;

                if (!_didIteratorError4) {
                  _context.next = 98;
                  break;
                }

                throw _iteratorError4;

              case 98:
                return _context.finish(95);

              case 99:
                return _context.finish(92);

              case 100:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this, [[3, 17, 21, 29], [22,, 24, 28], [33, 37, 41, 49], [42,, 44, 48], [52, 56, 60, 68], [61,, 63, 67], [72, 88, 92, 100], [93,, 95, 99]]);
      }));

      function updateRemoteRefs(_x) {
        return _ref2.apply(this, arguments);
      }

      return updateRemoteRefs;
    }()
  }, {
    key: 'resolve',
    value: function () {
      var _ref12 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref11) {
        var gitdir = _ref11.gitdir,
            ref = _ref11.ref,
            depth = _ref11.depth;
        var sha, text, candidates;
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (!(depth !== undefined)) {
                  _context2.next = 4;
                  break;
                }

                depth--;

                if (!(depth === -1)) {
                  _context2.next = 4;
                  break;
                }

                return _context2.abrupt('return', ref);

              case 4:
                sha = void 0;
                // Is it a ref pointer?

                if (!ref.startsWith('ref: ')) {
                  _context2.next = 8;
                  break;
                }

                ref = ref.slice('ref: '.length);
                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: ref, depth: depth }));

              case 8:
                if (!(ref.length === 40 && /[0-9a-f]{40}/.test(ref))) {
                  _context2.next = 10;
                  break;
                }

                return _context2.abrupt('return', ref);

              case 10:
                if (!(ref === 'HEAD' || ref === 'MERGE_HEAD')) {
                  _context2.next = 16;
                  break;
                }

                _context2.next = 13;
                return read(gitdir + '/' + ref, { encoding: 'utf8' });

              case 13:
                sha = _context2.sent;

                if (!sha) {
                  _context2.next = 16;
                  break;
                }

                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

              case 16:
                if (!ref.startsWith('refs/')) {
                  _context2.next = 22;
                  break;
                }

                _context2.next = 19;
                return read(gitdir + '/' + ref, { encoding: 'utf8' });

              case 19:
                sha = _context2.sent;

                if (!sha) {
                  _context2.next = 22;
                  break;
                }

                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

              case 22:
                _context2.next = 24;
                return read(gitdir + '/refs/heads/' + ref, { encoding: 'utf8' });

              case 24:
                sha = _context2.sent;

                if (!sha) {
                  _context2.next = 27;
                  break;
                }

                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

              case 27:
                _context2.next = 29;
                return read(gitdir + '/refs/tags/' + ref, { encoding: 'utf8' });

              case 29:
                sha = _context2.sent;

                if (!sha) {
                  _context2.next = 32;
                  break;
                }

                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

              case 32:
                _context2.next = 34;
                return read(gitdir + '/refs/remotes/' + ref, { encoding: 'utf8' });

              case 34:
                sha = _context2.sent;

                if (!sha) {
                  _context2.next = 37;
                  break;
                }

                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

              case 37:
                _context2.next = 39;
                return read(gitdir + '/packed-refs', { encoding: 'utf8' });

              case 39:
                text = _context2.sent;

                if (!(text && text.includes(ref))) {
                  _context2.next = 49;
                  break;
                }

                candidates = text.trim().split('\n').filter(function (x) {
                  return x.endsWith(ref);
                }).filter(function (x) {
                  return !x.startsWith('#');
                });

                if (!(candidates.length > 1)) {
                  _context2.next = 46;
                  break;
                }

                throw new Error('Could not resolve ambiguous reference ' + ref);

              case 46:
                if (!(candidates.length === 1)) {
                  _context2.next = 49;
                  break;
                }

                sha = candidates[0].split(' ')[0];
                return _context2.abrupt('return', GitRefManager.resolve({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

              case 49:
                throw new Error('Could not resolve reference ' + ref);

              case 50:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function resolve(_x2) {
        return _ref12.apply(this, arguments);
      }

      return resolve;
    }()
  }]);

  return GitRefManager;
}();

// @flow
function basicAuth(auth) {
  return 'Basic ' + buffer.Buffer.from(auth.username + ':' + auth.password).toString('base64');
}

var GitRemoteHTTP = function () {
  /*::
  GIT_URL : string
  refs : Map<string, string>
  symrefs : Map<string, string>
  capabilities : Set<string>
  auth : { username : string, password : string }
  */
  function GitRemoteHTTP(url /*: string */) {
    _classCallCheck(this, GitRemoteHTTP);

    // Auto-append the (necessary) .git if it's missing.
    if (!url.endsWith('.git')) url = url += '.git';
    this.GIT_URL = url;
  }

  _createClass(GitRemoteHTTP, [{
    key: 'preparePull',
    value: function () {
      var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee() {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return this.discover('git-upload-pack');

              case 2:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function preparePull() {
        return _ref.apply(this, arguments);
      }

      return preparePull;
    }()
  }, {
    key: 'preparePush',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return this.discover('git-receive-pack');

              case 2:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function preparePush() {
        return _ref2.apply(this, arguments);
      }

      return preparePush;
    }()
  }, {
    key: 'discover',
    value: function () {
      var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(service /*: string */) {
        var _this = this;

        var headers, res, data, read, lineOne, lineTwo, _lineTwo$toString$tri, _lineTwo$toString$tri2, firstRef, capabilities, _firstRef$split, _firstRef$split2, ref, name, line, _line$toString$trim$s, _line$toString$trim$s2, _ref4, _name, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, cap, m;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                this.capabilities = new _Set();
                this.refs = new _Map();
                this.symrefs = new _Map();
                headers = {};
                // headers['Accept'] = `application/x-${service}-advertisement`

                if (this.auth) {
                  headers['Authorization'] = basicAuth(this.auth);
                }
                _context3.next = 7;
                return pify(simpleGet)({
                  method: 'GET',
                  url: this.GIT_URL + '/info/refs?service=' + service,
                  headers: headers
                });

              case 7:
                res = _context3.sent;

                if (!(res.statusCode !== 200)) {
                  _context3.next = 10;
                  break;
                }

                throw new Error('Bad status code from server: ' + res.statusCode);

              case 10:
                _context3.next = 12;
                return pify(concat)(res);

              case 12:
                data = _context3.sent;

                // There is probably a better way to do this, but for now
                // let's just throw the result parser inline here.
                read = GitPktLine.reader(data);
                lineOne = read();
                // skip past any flushes

                while (lineOne === null) {
                  lineOne = read();
                }
                if (!(lineOne === true)) {
                  _context3.next = 18;
                  break;
                }

                throw new Error('Bad response from git server.');

              case 18:
                if (!(lineOne.toString('utf8') !== '# service=' + service + '\n')) {
                  _context3.next = 20;
                  break;
                }

                throw new Error('Expected \'# service=' + service + '\\n\' but got \'' + lineOne.toString('utf8') + '\'');

              case 20:
                lineTwo = read();
                // skip past any flushes

                while (lineTwo === null) {
                  lineTwo = read();
                } // In the edge case of a brand new repo, zero refs (and zero capabilities)
                // are returned.

                if (!(lineTwo === true)) {
                  _context3.next = 24;
                  break;
                }

                return _context3.abrupt('return');

              case 24:
                _lineTwo$toString$tri = lineTwo.toString('utf8').trim().split('\0'), _lineTwo$toString$tri2 = _slicedToArray(_lineTwo$toString$tri, 2), firstRef = _lineTwo$toString$tri2[0], capabilities = _lineTwo$toString$tri2[1];

                capabilities.split(' ').map(function (x) {
                  return _this.capabilities.add(x);
                });
                _firstRef$split = firstRef.split(' '), _firstRef$split2 = _slicedToArray(_firstRef$split, 2), ref = _firstRef$split2[0], name = _firstRef$split2[1];

                this.refs.set(name, ref);

              case 28:
                

                line = read();

                if (!(line === true)) {
                  _context3.next = 32;
                  break;
                }

                return _context3.abrupt('break', 35);

              case 32:
                if (line !== null) {
                  _line$toString$trim$s = line.toString('utf8').trim().split(' '), _line$toString$trim$s2 = _slicedToArray(_line$toString$trim$s, 2), _ref4 = _line$toString$trim$s2[0], _name = _line$toString$trim$s2[1];

                  this.refs.set(_name, _ref4);
                }
                _context3.next = 28;
                break;

              case 35:
                // Symrefs are thrown into the "capabilities" unfortunately.
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context3.prev = 38;
                for (_iterator = _getIterator(this.capabilities); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                  cap = _step.value;

                  if (cap.startsWith('symref=')) {
                    m = cap.match(/symref=([^:]+):(.*)/);

                    if (m.length === 3) {
                      this.symrefs.set(m[1], m[2]);
                    }
                  }
                }
                _context3.next = 46;
                break;

              case 42:
                _context3.prev = 42;
                _context3.t0 = _context3['catch'](38);
                _didIteratorError = true;
                _iteratorError = _context3.t0;

              case 46:
                _context3.prev = 46;
                _context3.prev = 47;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 49:
                _context3.prev = 49;

                if (!_didIteratorError) {
                  _context3.next = 52;
                  break;
                }

                throw _iteratorError;

              case 52:
                return _context3.finish(49);

              case 53:
                return _context3.finish(46);

              case 54:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this, [[38, 42, 46, 54], [47,, 49, 53]]);
      }));

      function discover(_x) {
        return _ref3.apply(this, arguments);
      }

      return discover;
    }()
  }, {
    key: 'push',
    value: function () {
      var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee4(stream$$1 /*: ReadableStream */) {
        var service, res;
        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                service = 'git-receive-pack';
                _context4.next = 3;
                return this.stream({ stream: stream$$1, service: service });

              case 3:
                res = _context4.sent;
                return _context4.abrupt('return', res);

              case 5:
              case 'end':
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function push(_x2) {
        return _ref5.apply(this, arguments);
      }

      return push;
    }()
  }, {
    key: 'pull',
    value: function () {
      var _ref6 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee5(stream$$1 /*: ReadableStream */) {
        var service, res;
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                service = 'git-upload-pack';
                _context5.next = 3;
                return this.stream({ stream: stream$$1, service: service });

              case 3:
                res = _context5.sent;
                return _context5.abrupt('return', res);

              case 5:
              case 'end':
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function pull(_x3) {
        return _ref6.apply(this, arguments);
      }

      return pull;
    }()
  }, {
    key: 'stream',
    value: function () {
      var _ref8 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee7(_ref7) {
        var _stream = _ref7.stream,
            service = _ref7.service;
        var headers, res, read, packetlines, packfile, progress, nextBit;
        return _regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                headers = {};

                headers['content-type'] = 'application/x-' + service + '-request';
                headers['accept'] = 'application/x-' + service + '-result';
                headers['user-agent'] = 'git/' + name + '@' + version;
                if (this.auth) {
                  headers['authorization'] = basicAuth(this.auth);
                }
                _context7.next = 7;
                return pify(simpleGet)({
                  method: 'POST',
                  url: this.GIT_URL + '/' + service,
                  body: _stream,
                  headers: headers
                });

              case 7:
                res = _context7.sent;

                if (!(service === 'git-receive-pack')) {
                  _context7.next = 10;
                  break;
                }

                return _context7.abrupt('return', res);

              case 10:
                // Parse the response!
                read = GitPktLine.streamReader(res);
                // And now for the ridiculous side-band-64k protocol

                packetlines = new stream.PassThrough();
                packfile = new stream.PassThrough();
                progress = new stream.PassThrough();
                // TODO: Use a proper through stream?

                nextBit = function () {
                  var _ref9 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee6() {
                    var line, error;
                    return _regeneratorRuntime.wrap(function _callee6$(_context6) {
                      while (1) {
                        switch (_context6.prev = _context6.next) {
                          case 0:
                            _context6.next = 2;
                            return read();

                          case 2:
                            line = _context6.sent;

                            if (!(line === null)) {
                              _context6.next = 8;
                              break;
                            }

                            packetlines.end();
                            progress.end();
                            packfile.end();
                            return _context6.abrupt('return');

                          case 8:
                            _context6.t0 = line[0];
                            _context6.next = _context6.t0 === 1 ? 11 : _context6.t0 === 2 ? 13 : _context6.t0 === 3 ? 15 : 19;
                            break;

                          case 11:
                            // pack data
                            packfile.write(line.slice(1));
                            return _context6.abrupt('break', 20);

                          case 13:
                            // progress message
                            progress.write(line.slice(1));
                            return _context6.abrupt('break', 20);

                          case 15:
                            // fatal error message just before stream aborts
                            error = line.slice(1);

                            progress.write(error);
                            packfile.destroy(new Error(error.toString('utf8')));
                            return _context6.abrupt('return');

                          case 19:
                            // Not part of the side-band-64k protocol
                            packetlines.write(line.slice(0));

                          case 20:
                            // Careful not to blow up the stack.
                            // I think Promises in a tail-call position should be OK.
                            nextBit();

                          case 21:
                          case 'end':
                            return _context6.stop();
                        }
                      }
                    }, _callee6, this);
                  }));

                  return function nextBit() {
                    return _ref9.apply(this, arguments);
                  };
                }();

                nextBit();
                return _context7.abrupt('return', {
                  packetlines: packetlines,
                  packfile: packfile,
                  progress: progress
                });

              case 17:
              case 'end':
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function stream$$1(_x4) {
        return _ref8.apply(this, arguments);
      }

      return stream$$1;
    }() /*: {
        stream: ReadableStream,
        service: string
        } */

  }]);

  return GitRemoteHTTP;
}();



var managers = Object.freeze({
	GitConfigManager: GitConfigManager,
	GitShallowManager: GitShallowManager,
	GitIndexManager: GitIndexManager,
	GitIgnoreManager: GitIgnoreManager,
	GitObjectManager: GitObjectManager,
	GitRefManager: GitRefManager,
	GitRemoteHTTP: GitRemoteHTTP
});

var add = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var gitdir = _ref.gitdir,
        workdir = _ref.workdir,
        filepath = _ref.filepath,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    var type, object, oid;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            type = 'blob';
            _context2.next = 4;
            return read(path.join(workdir, filepath));

          case 4:
            object = _context2.sent;

            if (!(object === null)) {
              _context2.next = 7;
              break;
            }

            throw new Error('Could not read file \'' + filepath + '\'');

          case 7:
            _context2.next = 9;
            return GitObjectManager.write({ gitdir: gitdir, type: type, object: object });

          case 9:
            oid = _context2.sent;
            _context2.next = 12;
            return GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(index) {
                var stats;
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        _context.next = 2;
                        return pify(fs$$1.lstat)(path.join(workdir, filepath));

                      case 2:
                        stats = _context.sent;

                        index.insert({ filepath: filepath, stats: stats, oid: oid });

                      case 4:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }());

          case 12:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function add(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var init = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    var folders;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            folders = ['hooks', 'info', 'objects/info', 'objects/pack', 'refs/heads', 'refs/tags'];

            folders = folders.map(function (dir) {
              return gitdir + '/' + dir;
            });
            _context.next = 5;
            return mkdirs(folders);

          case 5:
            _context.next = 7;
            return write(gitdir + '/config', '[core]\n' + '\trepositoryformatversion = 0\n' + '\tfilemode = false\n' + '\tbare = false\n' + '\tlogallrefupdates = true\n' + '\tsymlinks = false\n' + '\tignorecase = true\n');

          case 7:
            _context.next = 9;
            return write(gitdir + '/HEAD', 'ref: refs/heads/master\n');

          case 9:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function init(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var config = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(args) {
    var gitdir, path$$1, value, fs$$1, config, _value;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            gitdir = args.gitdir, path$$1 = args.path, value = args.value, fs$$1 = args.fs;

            fs$$1 = fs$$1 || defaultfs();
            setfs(fs$$1);
            _context.next = 5;
            return GitConfigManager.get({ gitdir: gitdir });

          case 5:
            config = _context.sent;

            if (!(value === undefined && !args.hasOwnProperty('value'))) {
              _context.next = 13;
              break;
            }

            _context.next = 9;
            return config.get(path$$1);

          case 9:
            _value = _context.sent;
            return _context.abrupt('return', _value);

          case 13:
            _context.next = 15;
            return config.set(path$$1, value);

          case 15:
            _context.next = 17;
            return GitConfigManager.save({ gitdir: gitdir, config: config });

          case 17:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function config(_x) {
    return _ref.apply(this, arguments);
  };
}();

// @flow
/*::
import type {Writable} from 'stream'
*/

var types = {
  1: 'commit',
  2: 'tree',
  3: 'blob',
  4: 'tag',
  6: 'ofs-delta',
  7: 'ref-delta'
};

function parseVarInt(buffer$$1 /*: Buffer */) {
  var n = 0;
  for (var i = 0; i < buffer$$1.byteLength; i++) {
    n = (buffer$$1[i] & 127) + (n << 7);
    if ((buffer$$1[i] & 128) === 0) {
      if (i !== buffer$$1.byteLength - 1) throw new Error('Invalid varint buffer');
      return n;
    }
  }
  throw new Error('Invalid varint buffer');
}

// TODO: Move this to 'plumbing'
var unpack = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref /*: {gitdir: string, inputStream: ReadableStream, onprogress: Function} */
  ) {
    var gitdir = _ref.gitdir,
        inputStream = _ref.inputStream,
        onprogress = _ref.onprogress,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            return _context2.abrupt('return', new _Promise(function (resolve, reject) {
              var _this = this;

              // Read header
              peek(inputStream, 12, function (err, data, inputStream) {
                if (err) return reject(err);
                var iden = data.slice(0, 4).toString('utf8');
                if (iden !== 'PACK') {
                  throw new Error('Packfile started with \'' + iden + '\'. Expected \'PACK\'');
                }
                var ver = data.slice(4, 8).toString('hex');
                if (ver !== '00000002') {
                  throw new Error('Unknown packfile version \'' + ver + '\'. Expected 00000002.');
                }
                // Read a 4 byte (32-bit) int
                var numObjects = data.readInt32BE(8);
                if (onprogress !== undefined) {
                  onprogress({ loaded: 0, total: numObjects, lengthComputable: true });
                }
                if (numObjects === 0) return resolve();
                // And on our merry way
                var totalTime = 0;
                var totalApplyDeltaTime = 0;
                var totalWriteFileTime = 0;
                var totalReadFileTime = 0;
                var offsetMap = new _Map();
                inputStream.pipe(listpack()).pipe(thru(function () {
                  var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref3, next) {
                    var data = _ref3.data,
                        type = _ref3.type,
                        reference = _ref3.reference,
                        offset = _ref3.offset,
                        num = _ref3.num;

                    var oid, _ref5, object, _type, result, newoid, absoluteOffset, referenceOid, _ref6, _type2, _object, _result, _oid, _oid2, perfentry;

                    return _regeneratorRuntime.wrap(function _callee$(_context) {
                      while (1) {
                        switch (_context.prev = _context.next) {
                          case 0:
                            type = types[type];
                            marky.mark(type + ' #' + num + ' ' + data.length + 'B');

                            if (!(type === 'ref-delta')) {
                              _context.next = 28;
                              break;
                            }

                            oid = buffer.Buffer.from(reference).toString('hex');
                            _context.prev = 4;

                            marky.mark('readFile');
                            _context.next = 8;
                            return GitObjectManager.read({
                              gitdir: gitdir,
                              oid: oid
                            });

                          case 8:
                            _ref5 = _context.sent;
                            object = _ref5.object;
                            _type = _ref5.type;

                            totalReadFileTime += marky.stop('readFile').duration;
                            marky.mark('applyDelta');
                            result = applyDelta(data, object);

                            totalApplyDeltaTime += marky.stop('applyDelta').duration;
                            marky.mark('writeFile');
                            _context.next = 18;
                            return GitObjectManager.write({
                              gitdir: gitdir,
                              type: _type,
                              object: result
                            });

                          case 18:
                            newoid = _context.sent;

                            totalWriteFileTime += marky.stop('writeFile').duration;
                            // console.log(`${type} ${newoid} ref-delta ${oid}`)
                            offsetMap.set(offset, newoid);
                            _context.next = 26;
                            break;

                          case 23:
                            _context.prev = 23;
                            _context.t0 = _context['catch'](4);
                            throw new Error('Could not find object ' + reference + ' ' + oid + ' that is referenced by a ref-delta object in packfile at byte offset ' + offset + '.');

                          case 26:
                            _context.next = 49;
                            break;

                          case 28:
                            if (!(type === 'ofs-delta')) {
                              _context.next = 43;
                              break;
                            }

                            // Note: this might be not working because offsets might not be
                            // guaranteed to be on object boundaries? In which case we'd need
                            // to write the packfile to disk first, I think.
                            // For now I've "solved" it by simply not advertising ofs-delta as a capability
                            // during the HTTP request, so Github will only send ref-deltas not ofs-deltas.
                            absoluteOffset = offset - parseVarInt(reference);
                            referenceOid = offsetMap.get(absoluteOffset);
                            // console.log(`${offset} ofs-delta ${absoluteOffset} ${referenceOid}`)

                            _context.next = 33;
                            return GitObjectManager.read({
                              gitdir: gitdir,
                              oid: referenceOid
                            });

                          case 33:
                            _ref6 = _context.sent;
                            _type2 = _ref6.type;
                            _object = _ref6.object;
                            _result = applyDelta(data, _object);
                            _context.next = 39;
                            return GitObjectManager.write({
                              gitdir: gitdir,
                              type: _type2,
                              object: _result
                            });

                          case 39:
                            _oid = _context.sent;

                            // console.log(`${offset} ${type} ${oid} ofs-delta ${referenceOid}`)
                            offsetMap.set(offset, _oid);
                            _context.next = 49;
                            break;

                          case 43:
                            marky.mark('writeFile');
                            _context.next = 46;
                            return GitObjectManager.write({
                              gitdir: gitdir,
                              type: type,
                              object: data
                            });

                          case 46:
                            _oid2 = _context.sent;

                            totalWriteFileTime += marky.stop('writeFile').duration;
                            // console.log(`${offset} ${type} ${oid}`)
                            offsetMap.set(offset, _oid2);

                          case 49:
                            if (onprogress !== undefined) {
                              onprogress({
                                loaded: numObjects - num,
                                total: numObjects,
                                lengthComputable: true
                              });
                            }
                            perfentry = marky.stop(type + ' #' + num + ' ' + data.length + 'B');

                            totalTime += perfentry.duration;

                            if (!(num === 0)) {
                              _context.next = 58;
                              break;
                            }

                            console.log('Total time unpacking objects: ' + totalTime);
                            console.log('Total time applying deltas: ' + totalApplyDeltaTime);
                            console.log('Total time reading files: ' + totalReadFileTime);
                            console.log('Total time writing files: ' + totalWriteFileTime);
                            return _context.abrupt('return', resolve());

                          case 58:
                            next(null);

                          case 59:
                          case 'end':
                            return _context.stop();
                        }
                      }
                    }, _callee, _this, [[4, 23]]);
                  }));

                  return function (_x2, _x3) {
                    return _ref4.apply(this, arguments);
                  };
                }())).on('error', reject).on('finish', resolve);
              });
            }));

          case 2:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function unpack(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var fetchPackfile = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var _this = this;

    var gitdir = _ref.gitdir,
        ref = _ref.ref,
        remote = _ref.remote,
        url = _ref.url,
        authUsername = _ref.authUsername,
        authPassword = _ref.authPassword,
        _ref$depth = _ref.depth,
        depth = _ref$depth === undefined ? null : _ref$depth,
        _ref$since = _ref.since,
        since = _ref$since === undefined ? null : _ref$since,
        _ref$exclude = _ref.exclude,
        exclude = _ref$exclude === undefined ? [] : _ref$exclude,
        _ref$relative = _ref.relative,
        relative = _ref$relative === undefined ? false : _ref$relative,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var remoteHTTP, want, capabilities, packstream, oids, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, oid, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, x, have, response;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);

            if (!(depth !== null)) {
              _context2.next = 5;
              break;
            }

            if (!_Number$isNaN(parseInt(depth))) {
              _context2.next = 4;
              break;
            }

            throw new Error('Invalid value for depth argument: ' + depth);

          case 4:
            depth = parseInt(depth);

          case 5:
            remote = remote || 'origin';

            if (!(url === undefined)) {
              _context2.next = 10;
              break;
            }

            _context2.next = 9;
            return config({
              gitdir: gitdir,
              path: 'remote.' + remote + '.url'
            });

          case 9:
            url = _context2.sent;

          case 10:
            remoteHTTP = new GitRemoteHTTP(url);

            if (authUsername !== undefined && authPassword !== undefined) {
              remoteHTTP.auth = {
                username: authUsername,
                password: authPassword
              };
            }
            _context2.next = 14;
            return remoteHTTP.preparePull();

          case 14:
            if (!(depth !== null && !remoteHTTP.capabilities.has('shallow'))) {
              _context2.next = 16;
              break;
            }

            throw new Error('Remote does not support shallow fetches');

          case 16:
            if (!(since !== null && !remoteHTTP.capabilities.has('deepen-since'))) {
              _context2.next = 18;
              break;
            }

            throw new Error('Remote does not support shallow fetches by date');

          case 18:
            if (!(exclude.length > 0 && !remoteHTTP.capabilities.has('deepen-not'))) {
              _context2.next = 20;
              break;
            }

            throw new Error('Remote does not support shallow fetches excluding commits reachable by refs');

          case 20:
            if (!(relative === true && !remoteHTTP.capabilities.has('deepen-relative'))) {
              _context2.next = 22;
              break;
            }

            throw new Error('Remote does not support shallow fetches relative to the current shallow depth');

          case 22:
            _context2.next = 24;
            return GitRefManager.updateRemoteRefs({
              gitdir: gitdir,
              remote: remote,
              refs: remoteHTTP.refs,
              symrefs: remoteHTTP.symrefs
            });

          case 24:
            _context2.next = 26;
            return GitRefManager.resolve({
              gitdir: gitdir,
              ref: 'refs/remotes/' + remote + '/' + ref
            });

          case 26:
            want = _context2.sent;

            // Note: I removed "ofs-delta" from the capabilities list and now
            // Github uses all ref-deltas when I fetch packfiles instead of all ofs-deltas. Nice!
            capabilities = 'multi_ack_detailed no-done side-band-64k thin-pack agent=git/' + name + '@' + version + (relative ? ' deepen-relative' : '');
            packstream = new stream__default.PassThrough();

            packstream.write(GitPktLine.encode('want ' + want + ' ' + capabilities + '\n'));
            _context2.next = 32;
            return GitShallowManager.read({ gitdir: gitdir });

          case 32:
            oids = _context2.sent;

            if (!(oids.size > 0 && remoteHTTP.capabilities.has('shallow'))) {
              _context2.next = 53;
              break;
            }

            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context2.prev = 37;

            for (_iterator = _getIterator(oids); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              oid = _step.value;

              packstream.write(GitPktLine.encode('shallow ' + oid + '\n'));
            }
            _context2.next = 45;
            break;

          case 41:
            _context2.prev = 41;
            _context2.t0 = _context2['catch'](37);
            _didIteratorError = true;
            _iteratorError = _context2.t0;

          case 45:
            _context2.prev = 45;
            _context2.prev = 46;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 48:
            _context2.prev = 48;

            if (!_didIteratorError) {
              _context2.next = 51;
              break;
            }

            throw _iteratorError;

          case 51:
            return _context2.finish(48);

          case 52:
            return _context2.finish(45);

          case 53:
            if (depth !== null) {
              packstream.write(GitPktLine.encode('deepen ' + depth + '\n'));
            }
            if (since !== null) {
              packstream.write(GitPktLine.encode('deepen-since ' + Math.floor(since.valueOf() / 1000) + '\n'));
            }
            _iteratorNormalCompletion2 = true;
            _didIteratorError2 = false;
            _iteratorError2 = undefined;
            _context2.prev = 58;
            for (_iterator2 = _getIterator(exclude); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              x = _step2.value;

              packstream.write(GitPktLine.encode('deepen-not ' + x + '\n'));
            }
            _context2.next = 66;
            break;

          case 62:
            _context2.prev = 62;
            _context2.t1 = _context2['catch'](58);
            _didIteratorError2 = true;
            _iteratorError2 = _context2.t1;

          case 66:
            _context2.prev = 66;
            _context2.prev = 67;

            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }

          case 69:
            _context2.prev = 69;

            if (!_didIteratorError2) {
              _context2.next = 72;
              break;
            }

            throw _iteratorError2;

          case 72:
            return _context2.finish(69);

          case 73:
            return _context2.finish(66);

          case 74:
            packstream.write(GitPktLine.flush());
            have = null;
            _context2.prev = 76;
            _context2.next = 79;
            return GitRefManager.resolve({ gitdir: gitdir, ref: ref });

          case 79:
            have = _context2.sent;
            _context2.next = 84;
            break;

          case 82:
            _context2.prev = 82;
            _context2.t2 = _context2['catch'](76);

          case 84:
            if (have) {
              packstream.write(GitPktLine.encode('have ' + have + '\n'));
              packstream.write(GitPktLine.flush());
            }
            packstream.end(GitPktLine.encode('done\n'));
            _context2.next = 88;
            return remoteHTTP.pull(packstream);

          case 88:
            response = _context2.sent;

            response.packetlines.pipe(thru(function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(data, next) {
                var line, _oid, _oid2;

                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        line = data.toString('utf8');

                        if (!line.startsWith('shallow')) {
                          _context.next = 10;
                          break;
                        }

                        _oid = line.slice(-41).trim();

                        if (!(_oid.length !== 40)) {
                          _context.next = 5;
                          break;
                        }

                        throw new Error('non-40 character \'shallow\' oid: ' + _oid);

                      case 5:
                        oids.add(_oid);
                        _context.next = 8;
                        return GitShallowManager.write({ gitdir: gitdir, oids: oids });

                      case 8:
                        _context.next = 17;
                        break;

                      case 10:
                        if (!line.startsWith('unshallow')) {
                          _context.next = 17;
                          break;
                        }

                        _oid2 = line.slice(-41).trim();

                        if (!(_oid2.length !== 40)) {
                          _context.next = 14;
                          break;
                        }

                        throw new Error('non-40 character \'shallow\' oid: ' + _oid2);

                      case 14:
                        oids.delete(_oid2);
                        _context.next = 17;
                        return GitShallowManager.write({ gitdir: gitdir, oids: oids });

                      case 17:
                        next(null, data);

                      case 18:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, _this);
              }));

              return function (_x2, _x3) {
                return _ref3.apply(this, arguments);
              };
            }()));
            return _context2.abrupt('return', response);

          case 91:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this, [[37, 41, 45, 53], [46,, 48, 52], [58, 62, 66, 74], [67,, 69, 73], [76, 82]]);
  }));

  return function fetchPackfile(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var fetch = function () {
  var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref4) {
    var gitdir = _ref4.gitdir,
        _ref4$ref = _ref4.ref,
        ref = _ref4$ref === undefined ? 'HEAD' : _ref4$ref,
        remote = _ref4.remote,
        url = _ref4.url,
        authUsername = _ref4.authUsername,
        authPassword = _ref4.authPassword,
        depth = _ref4.depth,
        since = _ref4.since,
        exclude = _ref4.exclude,
        relative = _ref4.relative,
        onprogress = _ref4.onprogress,
        fs$$1 = _ref4.fs;
    var response;
    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            _context3.next = 2;
            return fetchPackfile({
              gitdir: gitdir,
              ref: ref,
              remote: remote,
              url: url,
              authUsername: authUsername,
              authPassword: authPassword,
              depth: depth,
              since: since,
              exclude: exclude,
              relative: relative,
              fs: fs$$1
            });

          case 2:
            response = _context3.sent;
            _context3.next = 5;
            return unpack({ gitdir: gitdir, inputStream: response.packfile, onprogress: onprogress });

          case 5:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function fetch(_x4) {
    return _ref5.apply(this, arguments);
  };
}();

var writeTreeToDisk = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        workdir = _ref.workdir,
        index = _ref.index,
        prefix = _ref.prefix,
        tree = _ref.tree,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry, _ref3, type, object, entrypath, filepath, stats, _tree;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 4;
            _iterator = _getIterator(tree);

          case 6:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 33;
              break;
            }

            entry = _step.value;
            _context.next = 10;
            return GitObjectManager.read({
              gitdir: gitdir,
              oid: entry.oid
            });

          case 10:
            _ref3 = _context.sent;
            type = _ref3.type;
            object = _ref3.object;
            entrypath = prefix === '' ? entry.path : prefix + '/' + entry.path;
            filepath = path.join(workdir, prefix, entry.path);
            _context.t0 = type;
            _context.next = _context.t0 === 'blob' ? 18 : _context.t0 === 'tree' ? 25 : 29;
            break;

          case 18:
            _context.next = 20;
            return write(filepath, object);

          case 20:
            _context.next = 22;
            return pify(fs$$1.lstat)(filepath);

          case 22:
            stats = _context.sent;

            index.insert({
              filepath: entrypath,
              stats: stats,
              oid: entry.oid
            });
            return _context.abrupt('break', 30);

          case 25:
            _tree = GitTree.from(object);
            _context.next = 28;
            return writeTreeToDisk({
              gitdir: gitdir,
              workdir: workdir,
              index: index,
              prefix: entrypath,
              tree: _tree
            });

          case 28:
            return _context.abrupt('break', 30);

          case 29:
            throw new Error('Unexpected object type ' + type + ' found in tree for \'' + entrypath + '\'');

          case 30:
            _iteratorNormalCompletion = true;
            _context.next = 6;
            break;

          case 33:
            _context.next = 39;
            break;

          case 35:
            _context.prev = 35;
            _context.t1 = _context['catch'](4);
            _didIteratorError = true;
            _iteratorError = _context.t1;

          case 39:
            _context.prev = 39;
            _context.prev = 40;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 42:
            _context.prev = 42;

            if (!_didIteratorError) {
              _context.next = 45;
              break;
            }

            throw _iteratorError;

          case 45:
            return _context.finish(42);

          case 46:
            return _context.finish(39);

          case 47:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[4, 35, 39, 47], [40,, 42, 46]]);
  }));

  return function writeTreeToDisk(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var checkout = function () {
  var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref4) {
    var workdir = _ref4.workdir,
        gitdir = _ref4.gitdir,
        remote = _ref4.remote,
        ref = _ref4.ref;

    var oid, remoteRef, commit, comm, sha, _ref6, type, object, tree;

    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            // Get tree oid
            oid = void 0;

            if (!remote) {
              _context3.next = 18;
              break;
            }

            remoteRef = void 0;

            if (!(ref === undefined)) {
              _context3.next = 10;
              break;
            }

            _context3.next = 6;
            return GitRefManager.resolve({
              gitdir: gitdir,
              ref: remote + '/HEAD',
              depth: 2
            });

          case 6:
            remoteRef = _context3.sent;

            ref = path.basename(remoteRef);
            _context3.next = 11;
            break;

          case 10:
            remoteRef = remote + '/' + ref;

          case 11:
            _context3.next = 13;
            return GitRefManager.resolve({ gitdir: gitdir, ref: remoteRef });

          case 13:
            oid = _context3.sent;
            _context3.next = 16;
            return write(gitdir + '/refs/heads/' + ref, oid + '\n');

          case 16:
            _context3.next = 23;
            break;

          case 18:
            if (!(ref === undefined)) {
              _context3.next = 20;
              break;
            }

            throw new Error('Cannot checkout ref "undefined"');

          case 20:
            _context3.next = 22;
            return GitRefManager.resolve({ gitdir: gitdir, ref: ref });

          case 22:
            oid = _context3.sent;

          case 23:
            _context3.next = 25;
            return GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 25:
            commit = _context3.sent;

            if (!(commit.type !== 'commit')) {
              _context3.next = 28;
              break;
            }

            throw new Error('Unexpected type: ' + commit.type);

          case 28:
            comm = GitCommit.from(commit.object.toString('utf8'));
            sha = comm.headers().tree;
            // Get top-level tree

            _context3.next = 32;
            return GitObjectManager.read({ gitdir: gitdir, oid: sha });

          case 32:
            _ref6 = _context3.sent;
            type = _ref6.type;
            object = _ref6.object;

            if (!(type !== 'tree')) {
              _context3.next = 37;
              break;
            }

            throw new Error('Unexpected type: ' + type);

          case 37:
            tree = GitTree.from(object);
            // Acquire a lock on the index

            _context3.next = 40;
            return GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref7 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(index) {
                var _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, entry;

                return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                  while (1) {
                    switch (_context2.prev = _context2.next) {
                      case 0:
                        // TODO: Big optimization possible here.
                        // Instead of deleting and rewriting everything, only delete files
                        // that are not present in the new branch, and only write files that
                        // are not in the index or are in the index but have the wrong SHA.
                        _iteratorNormalCompletion2 = true;
                        _didIteratorError2 = false;
                        _iteratorError2 = undefined;
                        _context2.prev = 3;
                        _iterator2 = _getIterator(index);

                      case 5:
                        if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
                          _context2.next = 17;
                          break;
                        }

                        entry = _step2.value;
                        _context2.prev = 7;
                        _context2.next = 10;
                        return rm(path.join(workdir, entry.path));

                      case 10:
                        _context2.next = 14;
                        break;

                      case 12:
                        _context2.prev = 12;
                        _context2.t0 = _context2['catch'](7);

                      case 14:
                        _iteratorNormalCompletion2 = true;
                        _context2.next = 5;
                        break;

                      case 17:
                        _context2.next = 23;
                        break;

                      case 19:
                        _context2.prev = 19;
                        _context2.t1 = _context2['catch'](3);
                        _didIteratorError2 = true;
                        _iteratorError2 = _context2.t1;

                      case 23:
                        _context2.prev = 23;
                        _context2.prev = 24;

                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                          _iterator2.return();
                        }

                      case 26:
                        _context2.prev = 26;

                        if (!_didIteratorError2) {
                          _context2.next = 29;
                          break;
                        }

                        throw _iteratorError2;

                      case 29:
                        return _context2.finish(26);

                      case 30:
                        return _context2.finish(23);

                      case 31:
                        index.clear();
                        // Write files. TODO: Write them atomically
                        _context2.next = 34;
                        return writeTreeToDisk({ gitdir: gitdir, workdir: workdir, index: index, prefix: '', tree: tree });

                      case 34:
                        // Update HEAD TODO: Handle non-branch cases
                        write(gitdir + '/HEAD', 'ref: refs/heads/' + ref);

                      case 35:
                      case 'end':
                        return _context2.stop();
                    }
                  }
                }, _callee2, this, [[3, 19, 23, 31], [7, 12], [24,, 26, 30]]);
              }));

              return function (_x3) {
                return _ref7.apply(this, arguments);
              };
            }());

          case 40:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function checkout(_x2) {
    return _ref5.apply(this, arguments);
  };
}();

var clone = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var workdir = _ref.workdir,
        gitdir = _ref.gitdir,
        url = _ref.url,
        remote = _ref.remote,
        ref = _ref.ref,
        authUsername = _ref.authUsername,
        authPassword = _ref.authPassword,
        depth = _ref.depth,
        since = _ref.since,
        exclude = _ref.exclude,
        relative = _ref.relative,
        onprogress = _ref.onprogress,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            remote = remote || 'origin';
            _context.next = 4;
            return init({ gitdir: gitdir });

          case 4:
            _context.next = 6;
            return config({
              gitdir: gitdir,
              path: 'remote.' + remote + '.url',
              value: url
            });

          case 6:
            _context.next = 8;
            return fetch({
              gitdir: gitdir,
              ref: ref,
              remote: remote,
              authUsername: authUsername,
              authPassword: authPassword,
              depth: depth,
              since: since,
              exclude: exclude,
              relative: relative,
              onprogress: onprogress
            });

          case 8:
            _context.next = 10;
            return checkout({
              workdir: workdir,
              gitdir: gitdir,
              ref: ref,
              remote: remote
            });

          case 10:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function clone(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var constructTree = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        inode = _ref.inode;

    var children, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, _inode, entries, tree, oid;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            // use depth first traversal
            children = inode.children;
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 4;
            _iterator = _getIterator(children);

          case 6:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 16;
              break;
            }

            _inode = _step.value;

            if (!(_inode.type === 'tree')) {
              _context.next = 13;
              break;
            }

            _inode.metadata.mode = '040000';
            _context.next = 12;
            return constructTree({ gitdir: gitdir, inode: _inode });

          case 12:
            _inode.metadata.oid = _context.sent;

          case 13:
            _iteratorNormalCompletion = true;
            _context.next = 6;
            break;

          case 16:
            _context.next = 22;
            break;

          case 18:
            _context.prev = 18;
            _context.t0 = _context['catch'](4);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 22:
            _context.prev = 22;
            _context.prev = 23;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 25:
            _context.prev = 25;

            if (!_didIteratorError) {
              _context.next = 28;
              break;
            }

            throw _iteratorError;

          case 28:
            return _context.finish(25);

          case 29:
            return _context.finish(22);

          case 30:
            entries = children.map(function (inode) {
              return {
                mode: inode.metadata.mode,
                path: inode.basename,
                oid: inode.metadata.oid,
                type: inode.type
              };
            });
            tree = GitTree.from(entries);
            _context.next = 34;
            return GitObjectManager.write({
              gitdir: gitdir,
              type: 'tree',
              object: tree.toObject()
            });

          case 34:
            oid = _context.sent;
            return _context.abrupt('return', oid);

          case 36:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[4, 18, 22, 30], [23,, 25, 29]]);
  }));

  return function constructTree(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var commit = function () {
  var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref3) {
    var gitdir = _ref3.gitdir,
        author = _ref3.author,
        committer = _ref3.committer,
        message$$1 = _ref3.message,
        privateKeys = _ref3.privateKeys,
        _ref3$fs = _ref3.fs,
        fs$$1 = _ref3$fs === undefined ? defaultfs() : _ref3$fs;
    var authorDateTime, committerDateTime, oid;
    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            setfs(fs$$1);
            // Fill in missing arguments with default values
            if (author === undefined) author = {};

            if (!(author.name === undefined)) {
              _context3.next = 6;
              break;
            }

            _context3.next = 5;
            return config({ gitdir: gitdir, path: 'user.name' });

          case 5:
            author.name = _context3.sent;

          case 6:
            if (!(author.email === undefined)) {
              _context3.next = 10;
              break;
            }

            _context3.next = 9;
            return config({ gitdir: gitdir, path: 'user.email' });

          case 9:
            author.email = _context3.sent;

          case 10:
            committer = committer || author;
            authorDateTime = author.date || new Date();
            committerDateTime = committer.date || authorDateTime;
            oid = void 0;
            _context3.next = 16;
            return GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(index) {
                var inode, treeRef, parents, parent, comm, branch;
                return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                  while (1) {
                    switch (_context2.prev = _context2.next) {
                      case 0:
                        inode = flatFileListToDirectoryStructure(index.entries);
                        _context2.next = 3;
                        return constructTree({ gitdir: gitdir, inode: inode });

                      case 3:
                        treeRef = _context2.sent;
                        parents = void 0;
                        _context2.prev = 5;
                        _context2.next = 8;
                        return GitRefManager.resolve({ gitdir: gitdir, ref: 'HEAD' });

                      case 8:
                        parent = _context2.sent;

                        parents = [parent];
                        _context2.next = 15;
                        break;

                      case 12:
                        _context2.prev = 12;
                        _context2.t0 = _context2['catch'](5);

                        // Probably an initial commit
                        parents = [];

                      case 15:
                        comm = GitCommit.from({
                          tree: treeRef,
                          parent: parents,
                          author: {
                            name: author.name,
                            email: author.email,
                            timestamp: author.timestamp || Math.floor(authorDateTime.valueOf() / 1000),
                            timezoneOffset: author.timezoneOffset || 0
                          },
                          committer: {
                            name: committer.name,
                            email: committer.email,
                            timestamp: committer.timestamp || Math.floor(committerDateTime.valueOf() / 1000),
                            timezoneOffset: committer.timezoneOffset || 0
                          },
                          message: message$$1
                        });

                        if (!privateKeys) {
                          _context2.next = 20;
                          break;
                        }

                        _context2.next = 19;
                        return comm.sign(privateKeys);

                      case 19:
                        comm = _context2.sent;

                      case 20:
                        _context2.next = 22;
                        return GitObjectManager.write({
                          gitdir: gitdir,
                          type: 'commit',
                          object: comm.toObject()
                        });

                      case 22:
                        oid = _context2.sent;
                        _context2.next = 25;
                        return GitRefManager.resolve({
                          gitdir: gitdir,
                          ref: 'HEAD',
                          depth: 2
                        });

                      case 25:
                        branch = _context2.sent;
                        _context2.next = 28;
                        return write(path.join(gitdir, branch), oid + '\n');

                      case 28:
                      case 'end':
                        return _context2.stop();
                    }
                  }
                }, _callee2, this, [[5, 12]]);
              }));

              return function (_x3) {
                return _ref5.apply(this, arguments);
              };
            }());

          case 16:
            return _context3.abrupt('return', oid);

          case 17:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function commit(_x2) {
    return _ref4.apply(this, arguments);
  };
}();

var list = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var gitdir = _ref.gitdir,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    var filenames;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            filenames = void 0;
            _context2.next = 4;
            return GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(index) {
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        filenames = index.entries.map(function (x) {
                          return x.path;
                        });

                      case 1:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }());

          case 4:
            return _context2.abrupt('return', filenames);

          case 5:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function list(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// Note: this function will get more interesting once we add support for
// packed-refs.
var listBranches = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    var files, text, refs;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            _context.next = 3;
            return readDirAsFlatFileList(gitdir + '/refs/heads');

          case 3:
            files = _context.sent;

            files = files.map(function (x) {
              return x.replace(gitdir + '/refs/heads/', '');
            });
            _context.next = 7;
            return read(gitdir + '/packed-refs', { encoding: 'utf8' });

          case 7:
            text = _context.sent;

            if (text) {
              refs = text.trim().split('\n').filter(function (x) {
                return x.includes('refs/heads');
              }).map(function (x) {
                return x.replace(/^.+ refs\/heads\//, '').trim();
              }).filter(function (x) {
                return !files.includes(x);
              }); // remove duplicates

              files = files.concat(refs);
            }
            return _context.abrupt('return', files);

          case 10:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function listBranches(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var listCommits = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    /*: Set<string> */

    // Because git commits are named by their hash, there is no
    // way to construct a cycle. Therefore we won't worry about
    // setting a default recursion limit.
    var walk = function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(oid) {
        var _ref5, type, object, commit, parents, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                visited.add(oid);
                _context.next = 3;
                return GitObjectManager.read({ gitdir: gitdir, oid: oid });

              case 3:
                _ref5 = _context.sent;
                type = _ref5.type;
                object = _ref5.object;

                if (!(type !== 'commit')) {
                  _context.next = 8;
                  break;
                }

                throw new Error('Expected type commit but type is ' + type);

              case 8:
                commit = GitCommit.from(object);
                parents = commit.headers().parent;
                _iteratorNormalCompletion3 = true;
                _didIteratorError3 = false;
                _iteratorError3 = undefined;
                _context.prev = 13;
                _iterator3 = _getIterator(parents);

              case 15:
                if (_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done) {
                  _context.next = 23;
                  break;
                }

                oid = _step3.value;

                if (!(!finishingSet.has(oid) && !visited.has(oid))) {
                  _context.next = 20;
                  break;
                }

                _context.next = 20;
                return walk(oid);

              case 20:
                _iteratorNormalCompletion3 = true;
                _context.next = 15;
                break;

              case 23:
                _context.next = 29;
                break;

              case 25:
                _context.prev = 25;
                _context.t0 = _context['catch'](13);
                _didIteratorError3 = true;
                _iteratorError3 = _context.t0;

              case 29:
                _context.prev = 29;
                _context.prev = 30;

                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                  _iterator3.return();
                }

              case 32:
                _context.prev = 32;

                if (!_didIteratorError3) {
                  _context.next = 35;
                  break;
                }

                throw _iteratorError3;

              case 35:
                return _context.finish(32);

              case 36:
                return _context.finish(29);

              case 37:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this, [[13, 25, 29, 37], [30,, 32, 36]]);
      }));

      return function walk(_x2) {
        return _ref4.apply(this, arguments);
      };
    }();

    // Let's go walking!


    var gitdir = _ref.gitdir,
        start = _ref.start,
        finish = _ref.finish,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var startingSet, finishingSet, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, ref, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _ref3, _oid, visited, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, oid;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            startingSet = new _Set();
            finishingSet = new _Set();
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context2.prev = 6;
            _iterator = _getIterator(start);

          case 8:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context2.next = 18;
              break;
            }

            ref = _step.value;
            _context2.t0 = startingSet;
            _context2.next = 13;
            return GitRefManager.resolve({ gitdir: gitdir, ref: ref });

          case 13:
            _context2.t1 = _context2.sent;

            _context2.t0.add.call(_context2.t0, _context2.t1);

          case 15:
            _iteratorNormalCompletion = true;
            _context2.next = 8;
            break;

          case 18:
            _context2.next = 24;
            break;

          case 20:
            _context2.prev = 20;
            _context2.t2 = _context2['catch'](6);
            _didIteratorError = true;
            _iteratorError = _context2.t2;

          case 24:
            _context2.prev = 24;
            _context2.prev = 25;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 27:
            _context2.prev = 27;

            if (!_didIteratorError) {
              _context2.next = 30;
              break;
            }

            throw _iteratorError;

          case 30:
            return _context2.finish(27);

          case 31:
            return _context2.finish(24);

          case 32:
            _iteratorNormalCompletion2 = true;
            _didIteratorError2 = false;
            _iteratorError2 = undefined;
            _context2.prev = 35;
            _iterator2 = _getIterator(finish);

          case 37:
            if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
              _context2.next = 51;
              break;
            }

            _ref3 = _step2.value;
            _context2.prev = 39;
            _context2.next = 42;
            return GitRefManager.resolve({ gitdir: gitdir, ref: _ref3 });

          case 42:
            _oid = _context2.sent;

            finishingSet.add(_oid);
            _context2.next = 48;
            break;

          case 46:
            _context2.prev = 46;
            _context2.t3 = _context2['catch'](39);

          case 48:
            _iteratorNormalCompletion2 = true;
            _context2.next = 37;
            break;

          case 51:
            _context2.next = 57;
            break;

          case 53:
            _context2.prev = 53;
            _context2.t4 = _context2['catch'](35);
            _didIteratorError2 = true;
            _iteratorError2 = _context2.t4;

          case 57:
            _context2.prev = 57;
            _context2.prev = 58;

            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }

          case 60:
            _context2.prev = 60;

            if (!_didIteratorError2) {
              _context2.next = 63;
              break;
            }

            throw _iteratorError2;

          case 63:
            return _context2.finish(60);

          case 64:
            return _context2.finish(57);

          case 65:
            visited = new _Set();
            _iteratorNormalCompletion4 = true;
            _didIteratorError4 = false;
            _iteratorError4 = undefined;
            _context2.prev = 69;
            _iterator4 = _getIterator(startingSet);

          case 71:
            if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
              _context2.next = 78;
              break;
            }

            oid = _step4.value;
            _context2.next = 75;
            return walk(oid);

          case 75:
            _iteratorNormalCompletion4 = true;
            _context2.next = 71;
            break;

          case 78:
            _context2.next = 84;
            break;

          case 80:
            _context2.prev = 80;
            _context2.t5 = _context2['catch'](69);
            _didIteratorError4 = true;
            _iteratorError4 = _context2.t5;

          case 84:
            _context2.prev = 84;
            _context2.prev = 85;

            if (!_iteratorNormalCompletion4 && _iterator4.return) {
              _iterator4.return();
            }

          case 87:
            _context2.prev = 87;

            if (!_didIteratorError4) {
              _context2.next = 90;
              break;
            }

            throw _iteratorError4;

          case 90:
            return _context2.finish(87);

          case 91:
            return _context2.finish(84);

          case 92:
            return _context2.abrupt('return', visited);

          case 93:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this, [[6, 20, 24, 32], [25,, 27, 31], [35, 53, 57, 65], [39, 46], [58,, 60, 64], [69, 80, 84, 92], [85,, 87, 91]]);
  }));

  return function listCommits(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var listObjects = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref /*: {
                                                                                gitdir: string,
                                                                                oids: Array<string>
                                                                                } */
  ) {

    // We don't do the purest simplest recursion, because we can
    // avoid reading Blob objects entirely since the Tree objects
    // tell us which oids are Blobs and which are Trees.
    var walk = function () {
      var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(oid) {
        var _ref4, type, object, commit, tree, _tree, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                visited.add(oid);
                _context.next = 3;
                return GitObjectManager.read({ gitdir: gitdir, oid: oid });

              case 3:
                _ref4 = _context.sent;
                type = _ref4.type;
                object = _ref4.object;

                if (!(type === 'commit')) {
                  _context.next = 13;
                  break;
                }

                commit = GitCommit.from(object);
                tree = commit.headers().tree;
                _context.next = 11;
                return walk(tree);

              case 11:
                _context.next = 43;
                break;

              case 13:
                if (!(type === 'tree')) {
                  _context.next = 43;
                  break;
                }

                _tree = GitTree.from(object);
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context.prev = 18;
                _iterator = _getIterator( /*: TreeEntry */_tree);

              case 20:
                if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                  _context.next = 29;
                  break;
                }

                entry = _step.value;

                visited.add(entry.oid);
                // only recurse for trees

                if (!(entry.type === 'tree')) {
                  _context.next = 26;
                  break;
                }

                _context.next = 26;
                return walk(entry.oid);

              case 26:
                _iteratorNormalCompletion = true;
                _context.next = 20;
                break;

              case 29:
                _context.next = 35;
                break;

              case 31:
                _context.prev = 31;
                _context.t0 = _context['catch'](18);
                _didIteratorError = true;
                _iteratorError = _context.t0;

              case 35:
                _context.prev = 35;
                _context.prev = 36;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 38:
                _context.prev = 38;

                if (!_didIteratorError) {
                  _context.next = 41;
                  break;
                }

                throw _iteratorError;

              case 41:
                return _context.finish(38);

              case 42:
                return _context.finish(35);

              case 43:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this, [[18, 31, 35, 43], [36,, 38, 42]]);
      }));

      return function walk(_x2) {
        return _ref3.apply(this, arguments);
      };
    }();

    // Let's go walking!


    var gitdir = _ref.gitdir,
        oids = _ref.oids,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var visited, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, oid;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            visited /*: Set<string> */ = new _Set();
            _iteratorNormalCompletion2 = true;
            _didIteratorError2 = false;
            _iteratorError2 = undefined;
            _context2.prev = 5;
            _iterator2 = _getIterator(oids);

          case 7:
            if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
              _context2.next = 14;
              break;
            }

            oid = _step2.value;
            _context2.next = 11;
            return walk(oid);

          case 11:
            _iteratorNormalCompletion2 = true;
            _context2.next = 7;
            break;

          case 14:
            _context2.next = 20;
            break;

          case 16:
            _context2.prev = 16;
            _context2.t0 = _context2['catch'](5);
            _didIteratorError2 = true;
            _iteratorError2 = _context2.t0;

          case 20:
            _context2.prev = 20;
            _context2.prev = 21;

            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }

          case 23:
            _context2.prev = 23;

            if (!_didIteratorError2) {
              _context2.next = 26;
              break;
            }

            throw _iteratorError2;

          case 26:
            return _context2.finish(23);

          case 27:
            return _context2.finish(20);

          case 28:
            return _context2.abrupt('return', visited);

          case 29:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this, [[5, 16, 20, 28], [21,, 23, 27]]);
  }));

  return function listObjects(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var log = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        _ref$ref = _ref.ref,
        ref = _ref$ref === undefined ? 'HEAD' : _ref$ref,
        depth = _ref.depth,
        since = _ref.since,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var sinceTimestamp, commits, start, _ref3, type, object, currentCommit, oid, gitobject, _gitobject, _type, _object;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            sinceTimestamp = since === undefined ? undefined : Math.floor(since.valueOf() / 1000);
            // TODO: In the future, we may want to have an API where we return a
            // async iterator that emits commits.

            commits = [];
            _context.next = 5;
            return GitRefManager.resolve({ gitdir: gitdir, ref: ref });

          case 5:
            start = _context.sent;
            _context.next = 8;
            return GitObjectManager.read({ gitdir: gitdir, oid: start });

          case 8:
            _ref3 = _context.sent;
            type = _ref3.type;
            object = _ref3.object;

            if (!(type !== 'commit')) {
              _context.next = 13;
              break;
            }

            throw new Error('The given ref ' + ref + ' did not resolve to a commit but to a ' + type);

          case 13:
            currentCommit = _extends({ oid: start }, GitCommit.from(object).parse());

            commits.push(currentCommit);

          case 15:
            

            if (!(depth !== undefined && commits.length === depth)) {
              _context.next = 18;
              break;
            }

            return _context.abrupt('break', 42);

          case 18:
            if (!(currentCommit.parent.length === 0)) {
              _context.next = 20;
              break;
            }

            return _context.abrupt('break', 42);

          case 20:
            oid = currentCommit.parent[0];
            gitobject = void 0;
            _context.prev = 22;
            _context.next = 25;
            return GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 25:
            gitobject = _context.sent;
            _context.next = 32;
            break;

          case 28:
            _context.prev = 28;
            _context.t0 = _context['catch'](22);

            commits.push({
              oid: oid,
              error: _context.t0
            });
            return _context.abrupt('break', 42);

          case 32:
            _gitobject = gitobject, _type = _gitobject.type, _object = _gitobject.object;

            if (!(_type !== 'commit')) {
              _context.next = 36;
              break;
            }

            commits.push({
              oid: oid,
              error: new Error('Invalid commit parent ' + oid + ' is of type ' + _type)
            });
            return _context.abrupt('break', 42);

          case 36:
            currentCommit = _extends({ oid: oid }, GitCommit.from(_object).parse());

            if (!(sinceTimestamp !== undefined && currentCommit.author.timestamp <= sinceTimestamp)) {
              _context.next = 39;
              break;
            }

            return _context.abrupt('break', 42);

          case 39:
            commits.push(currentCommit);
            _context.next = 15;
            break;

          case 42:
            return _context.abrupt('return', commits);

          case 43:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[22, 28]]);
  }));

  return function log(_x) {
    return _ref2.apply(this, arguments);
  };
}();

/*::
import type {Writable} from 'stream'
*/

var types$1 = {
  commit: 16,
  tree: 32,
  blob: 48,
  tag: 64
  // TODO: Move this to 'plumbing'
};var pack = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var oids = _ref.oids,
        gitdir = _ref.gitdir,
        outputStream = _ref.outputStream,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var hash, stream$$1, write, writeObject, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, oid, _ref4, type, object, digest;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            writeObject = function writeObject(_ref3) {
              var stype = _ref3.stype,
                  object = _ref3.object;

              var lastFour = void 0,
                  multibyte = void 0,
                  length = void 0;
              // Object type is encoded in bits 654
              var type = types$1[stype];
              if (type === undefined) throw new Error('Unrecognized type: ' + stype);
              // The length encoding get complicated.
              length = object.length;
              // Whether the next byte is part of the variable-length encoded number
              // is encoded in bit 7
              multibyte = length > 15 ? 128 : 0;
              // Last four bits of length is encoded in bits 3210
              lastFour = length & 15;
              // Discard those bits
              length = length >>> 4;
              // The first byte is then (1-bit multibyte?), (3-bit type), (4-bit least sig 4-bits of length)
              var byte = (multibyte | type | lastFour).toString(16);
              write(byte, 'hex');
              // Now we keep chopping away at length 7-bits at a time until its zero,
              // writing out the bytes in what amounts to little-endian order.
              while (multibyte) {
                multibyte = length > 127 ? 128 : 0;
                byte = multibyte | length & 127;
                write(pad(2, byte.toString(16), '0'), 'hex');
                length = length >>> 7;
              }
              // Lastly, we can compress and write the object.
              write(buffer.Buffer.from(pako.deflate(object)));
            };

            write = function write(chunk, enc) {
              stream$$1.write(chunk, enc);
              hash.update(chunk, enc);
            };

            setfs(fs$$1);
            hash = crypto.createHash('sha1');
            stream$$1 = outputStream;


            write('PACK');
            write('00000002', 'hex');
            // Write a 4 byte (32-bit) int
            write(pad(8, oids.length.toString(16), '0'), 'hex');
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 11;
            _iterator = _getIterator(oids);

          case 13:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 24;
              break;
            }

            oid = _step.value;
            _context.next = 17;
            return GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 17:
            _ref4 = _context.sent;
            type = _ref4.type;
            object = _ref4.object;

            writeObject({ write: write, object: object, stype: type });

          case 21:
            _iteratorNormalCompletion = true;
            _context.next = 13;
            break;

          case 24:
            _context.next = 30;
            break;

          case 26:
            _context.prev = 26;
            _context.t0 = _context['catch'](11);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 30:
            _context.prev = 30;
            _context.prev = 31;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 33:
            _context.prev = 33;

            if (!_didIteratorError) {
              _context.next = 36;
              break;
            }

            throw _iteratorError;

          case 36:
            return _context.finish(33);

          case 37:
            return _context.finish(30);

          case 38:
            // Write SHA1 checksum
            digest = hash.digest();

            stream$$1.end(digest);
            return _context.abrupt('return', stream$$1);

          case 41:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[11, 26, 30, 38], [31,, 33, 37]]);
  }));

  return function pack(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var push = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        ref = _ref.ref,
        remote = _ref.remote,
        url = _ref.url,
        authUsername = _ref.authUsername,
        authPassword = _ref.authPassword,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    var fullRef, oid, httpRemote, commits, objects, packstream, oldoid, response;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            // TODO: Figure out how pushing tags works. (This only works for branches.)
            remote = remote || 'origin';

            if (!(url === undefined)) {
              _context.next = 6;
              break;
            }

            _context.next = 5;
            return config({
              gitdir: gitdir,
              path: 'remote.' + remote + '.url'
            });

          case 5:
            url = _context.sent;

          case 6:
            fullRef = ref.startsWith('refs/') ? ref : 'refs/heads/' + ref;
            _context.next = 9;
            return GitRefManager.resolve({ gitdir: gitdir, ref: ref });

          case 9:
            oid = _context.sent;
            httpRemote = new GitRemoteHTTP(url);

            if (authUsername !== undefined && authPassword !== undefined) {
              httpRemote.auth = {
                username: authUsername,
                password: authPassword
              };
            }
            _context.next = 14;
            return httpRemote.preparePush();

          case 14:
            _context.next = 16;
            return listCommits({
              gitdir: gitdir,
              start: [oid],
              finish: httpRemote.refs.values(),
              fs: fs$$1
            });

          case 16:
            commits = _context.sent;
            _context.next = 19;
            return listObjects({ gitdir: gitdir, oids: commits, fs: fs$$1 });

          case 19:
            objects = _context.sent;
            packstream = new stream__default.PassThrough();
            oldoid = httpRemote.refs.get(fullRef) || '0000000000000000000000000000000000000000';

            packstream.write(GitPktLine.encode(oldoid + ' ' + oid + ' ' + fullRef + '\0 report-status\n'));
            packstream.write(GitPktLine.flush());
            pack({
              gitdir: gitdir,
              oids: [].concat(_toConsumableArray(objects)),
              outputStream: packstream
            });
            _context.next = 27;
            return httpRemote.push(packstream);

          case 27:
            response = _context.sent;
            return _context.abrupt('return', response);

          case 29:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function push(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var remove = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var gitdir = _ref.gitdir,
        filepath = _ref.filepath,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            _context2.next = 3;
            return GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(index) {
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        index.delete({ filepath: filepath });

                      case 1:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }());

          case 3:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function remove(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var HttpKeyServer = new openpgp.HKP();

var verify = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        ref = _ref.ref,
        publicKeys = _ref.publicKeys,
        _ref$fs = _ref.fs,
        fs$$1 = _ref$fs === undefined ? defaultfs() : _ref$fs;

    var oid, _ref3, type, object, commit, author, keys, keyArray, validity;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            setfs(fs$$1);
            _context.next = 3;
            return GitRefManager.resolve({ gitdir: gitdir, ref: ref });

          case 3:
            oid = _context.sent;
            _context.next = 6;
            return GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 6:
            _ref3 = _context.sent;
            type = _ref3.type;
            object = _ref3.object;

            if (!(type !== 'commit')) {
              _context.next = 11;
              break;
            }

            throw new Error('git.verify() was expecting a ref type \'commit\' but got type \'' + type + '\'');

          case 11:
            commit = GitCommit.from(object);
            author = commit.headers().author;
            _context.next = 15;
            return commit.listSigningKeys();

          case 15:
            keys = _context.sent;

            if (publicKeys) {
              _context.next = 21;
              break;
            }

            _context.next = 19;
            return _Promise.all(keys.map(function (id) {
              return HttpKeyServer.lookup({ keyId: id });
            }));

          case 19:
            keyArray = _context.sent;

            publicKeys = keyArray.join('\n');

          case 21:
            _context.next = 23;
            return commit.verify(publicKeys);

          case 23:
            validity = _context.sent;

            if (validity) {
              _context.next = 26;
              break;
            }

            return _context.abrupt('return', false);

          case 26:
            return _context.abrupt('return', { author: author, keys: keys });

          case 27:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function verify(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var getOidAtPath = function () {
  var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref2 /*: {
                                                                                gitdir: string,
                                                                                tree: GitTree,
                                                                                path: string|Array<string>
                                                                                } */
  ) {
    var gitdir = _ref2.gitdir,
        tree = _ref2.tree,
        path$$1 = _ref2.path;

    var dirname, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry, _ref4, type, object, _tree;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (typeof path$$1 === 'string') path$$1 = path$$1.split('/');
            dirname = path$$1.shift();
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 5;
            _iterator = _getIterator(tree);

          case 7:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 25;
              break;
            }

            entry = _step.value;

            if (!(entry.path === dirname)) {
              _context.next = 22;
              break;
            }

            if (!(path$$1.length === 0)) {
              _context.next = 12;
              break;
            }

            return _context.abrupt('return', entry.oid);

          case 12:
            _context.next = 14;
            return GitObjectManager.read({
              gitdir: gitdir,
              oid: entry.oid
            });

          case 14:
            _ref4 = _context.sent;
            type = _ref4.type;
            object = _ref4.object;

            if (!(type === 'tree')) {
              _context.next = 20;
              break;
            }

            _tree = GitTree.from(object);
            return _context.abrupt('return', getOidAtPath({ gitdir: gitdir, tree: _tree, path: path$$1 }));

          case 20:
            if (!(type === 'blob')) {
              _context.next = 22;
              break;
            }

            throw new Error('Blob found where tree expected.');

          case 22:
            _iteratorNormalCompletion = true;
            _context.next = 7;
            break;

          case 25:
            _context.next = 31;
            break;

          case 27:
            _context.prev = 27;
            _context.t0 = _context['catch'](5);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 31:
            _context.prev = 31;
            _context.prev = 32;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 34:
            _context.prev = 34;

            if (!_didIteratorError) {
              _context.next = 37;
              break;
            }

            throw _iteratorError;

          case 37:
            return _context.finish(34);

          case 38:
            return _context.finish(31);

          case 39:
            return _context.abrupt('return', null);

          case 40:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[5, 27, 31, 39], [32,, 34, 38]]);
  }));

  return function getOidAtPath(_x) {
    return _ref3.apply(this, arguments);
  };
}();

var getHeadTree = function () {
  var _ref6 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref5) {
    var gitdir = _ref5.gitdir;

    var oid, _ref7, cobject, commit, _ref8, tobject, tree;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return GitRefManager.resolve({ gitdir: gitdir, ref: 'HEAD' });

          case 2:
            oid = _context2.sent;
            _context2.next = 5;
            return GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 5:
            _ref7 = _context2.sent;
            cobject = _ref7.object;
            commit = GitCommit.from(cobject);
            _context2.next = 10;
            return GitObjectManager.read({
              gitdir: gitdir,
              oid: commit.parseHeaders().tree
            });

          case 10:
            _ref8 = _context2.sent;
            tobject = _ref8.object;
            tree = GitTree.from(tobject).entries();
            return _context2.abrupt('return', tree);

          case 14:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function getHeadTree(_x2) {
    return _ref6.apply(this, arguments);
  };
}();

// For now we're just diffing individual files


// @flow
/*::
import type { Stats } from 'fs'
import type { CacheEntry } from '../models/GitIndex'
*/

function cacheIsStale(_ref /*: {
                           entry: CacheEntry,
                           stats: Stats
                           } */
) {
  var entry = _ref.entry,
      stats = _ref.stats;

  // Comparison based on the description in Paragraph 4 of
  // https://www.kernel.org/pub/software/scm/git/docs/technical/racy-git.txt
  return entry.mode !== stats.mode || entry.mtime.valueOf() !== stats.mtime.valueOf() || entry.ctime.valueOf() !== stats.ctime.valueOf() || entry.uid !== stats.uid || entry.gid !== stats.gid || entry.ino !== stats.ino >> 0 || entry.size !== stats.size;
}

var status = function () {
  var _ref10 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee5(_ref9 /*: {
                                                                                  workdir: string,
                                                                                  gitdir: string,
                                                                                  pathname: string
                                                                                  } */
  ) {
    var _this = this;

    var workdir = _ref9.workdir,
        gitdir = _ref9.gitdir,
        pathname = _ref9.pathname,
        _ref9$fs = _ref9.fs,
        fs$$1 = _ref9$fs === undefined ? defaultfs() : _ref9$fs;

    var ignored, headTree, treeOid, indexEntry, stats, H, I, W, getWorkdirOid, workdirOid, _workdirOid, _workdirOid2;

    return _regeneratorRuntime.wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            setfs(fs$$1);
            _context5.next = 3;
            return GitIgnoreManager.isIgnored({ gitdir: gitdir, workdir: workdir, pathname: pathname });

          case 3:
            ignored = _context5.sent;

            if (!ignored) {
              _context5.next = 6;
              break;
            }

            return _context5.abrupt('return', 'ignored');

          case 6:
            _context5.next = 8;
            return getHeadTree({ gitdir: gitdir });

          case 8:
            headTree = _context5.sent;
            _context5.next = 11;
            return getOidAtPath({ gitdir: gitdir, tree: headTree, path: pathname });

          case 11:
            treeOid = _context5.sent;
            indexEntry = null;
            // Acquire a lock on the index

            _context5.next = 15;
            return GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref11 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(index) {
                var _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, entry;

                return _regeneratorRuntime.wrap(function _callee3$(_context3) {
                  while (1) {
                    switch (_context3.prev = _context3.next) {
                      case 0:
                        _iteratorNormalCompletion2 = true;
                        _didIteratorError2 = false;
                        _iteratorError2 = undefined;
                        _context3.prev = 3;
                        _iterator2 = _getIterator(index);

                      case 5:
                        if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
                          _context3.next = 13;
                          break;
                        }

                        entry = _step2.value;

                        if (!(entry.path === pathname)) {
                          _context3.next = 10;
                          break;
                        }

                        indexEntry = entry;
                        return _context3.abrupt('break', 13);

                      case 10:
                        _iteratorNormalCompletion2 = true;
                        _context3.next = 5;
                        break;

                      case 13:
                        _context3.next = 19;
                        break;

                      case 15:
                        _context3.prev = 15;
                        _context3.t0 = _context3['catch'](3);
                        _didIteratorError2 = true;
                        _iteratorError2 = _context3.t0;

                      case 19:
                        _context3.prev = 19;
                        _context3.prev = 20;

                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                          _iterator2.return();
                        }

                      case 22:
                        _context3.prev = 22;

                        if (!_didIteratorError2) {
                          _context3.next = 25;
                          break;
                        }

                        throw _iteratorError2;

                      case 25:
                        return _context3.finish(22);

                      case 26:
                        return _context3.finish(19);

                      case 27:
                      case 'end':
                        return _context3.stop();
                    }
                  }
                }, _callee3, this, [[3, 15, 19, 27], [20,, 22, 26]]);
              }));

              return function (_x4) {
                return _ref11.apply(this, arguments);
              };
            }());

          case 15:
            stats = null;
            _context5.prev = 16;
            _context5.next = 19;
            return pify(fs$$1.lstat)(path.join(workdir, pathname));

          case 19:
            stats = _context5.sent;
            _context5.next = 26;
            break;

          case 22:
            _context5.prev = 22;
            _context5.t0 = _context5['catch'](16);

            if (!(_context5.t0.code !== 'ENOENT')) {
              _context5.next = 26;
              break;
            }

            throw _context5.t0;

          case 26:
            H = treeOid !== null; // head

            I = indexEntry !== null; // index

            W = stats !== null; // working dir

            getWorkdirOid = function () {
              var _ref12 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee4() {
                var object, workdirOid;
                return _regeneratorRuntime.wrap(function _callee4$(_context4) {
                  while (1) {
                    switch (_context4.prev = _context4.next) {
                      case 0:
                        if (!(I && !cacheIsStale({ entry: indexEntry, stats: stats }))) {
                          _context4.next = 4;
                          break;
                        }

                        return _context4.abrupt('return', indexEntry.oid);

                      case 4:
                        _context4.next = 6;
                        return read(path.join(workdir, pathname));

                      case 6:
                        object = _context4.sent;
                        _context4.next = 9;
                        return GitObjectManager.hash({
                          gitdir: gitdir,
                          type: 'blob',
                          object: object
                        });

                      case 9:
                        workdirOid = _context4.sent;
                        return _context4.abrupt('return', workdirOid);

                      case 11:
                      case 'end':
                        return _context4.stop();
                    }
                  }
                }, _callee4, _this);
              }));

              return function getWorkdirOid() {
                return _ref12.apply(this, arguments);
              };
            }();

            if (!(!H && !W && !I)) {
              _context5.next = 32;
              break;
            }

            return _context5.abrupt('return', 'absent');

          case 32:
            if (!(!H && !W && I)) {
              _context5.next = 34;
              break;
            }

            return _context5.abrupt('return', '*absent');

          case 34:
            if (!(!H && W && !I)) {
              _context5.next = 36;
              break;
            }

            return _context5.abrupt('return', '*added');

          case 36:
            if (!(!H && W && I)) {
              _context5.next = 41;
              break;
            }

            _context5.next = 39;
            return getWorkdirOid();

          case 39:
            workdirOid = _context5.sent;
            return _context5.abrupt('return', workdirOid === indexEntry.oid ? 'added' : '*added');

          case 41:
            if (!(H && !W && !I)) {
              _context5.next = 43;
              break;
            }

            return _context5.abrupt('return', 'deleted');

          case 43:
            if (!(H && !W && I)) {
              _context5.next = 45;
              break;
            }

            return _context5.abrupt('return', treeOid === indexEntry.oid ? '*deleted' : '*deleted');

          case 45:
            if (!(H && W && !I)) {
              _context5.next = 50;
              break;
            }

            _context5.next = 48;
            return getWorkdirOid();

          case 48:
            _workdirOid = _context5.sent;
            return _context5.abrupt('return', _workdirOid === treeOid ? '*undeleted' : '*undeletemodified');

          case 50:
            if (!(H && W && I)) {
              _context5.next = 59;
              break;
            }

            _context5.next = 53;
            return getWorkdirOid();

          case 53:
            _workdirOid2 = _context5.sent;

            if (!(_workdirOid2 === treeOid)) {
              _context5.next = 58;
              break;
            }

            return _context5.abrupt('return', _workdirOid2 === indexEntry.oid ? 'unmodified' : '*unmodified');

          case 58:
            return _context5.abrupt('return', _workdirOid2 === indexEntry.oid ? 'modified' : '*modified');

          case 59:
          case 'end':
            return _context5.stop();
        }
      }
    }, _callee5, this, [[16, 22]]);
  }));

  return function status(_x3) {
    return _ref10.apply(this, arguments);
  };
}();

var test = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(filepath) {
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return pify(defaultfs().lstat)(path.join(filepath, '.git'));

          case 3:
            return _context.abrupt('return', true);

          case 6:
            _context.prev = 6;
            _context.t0 = _context['catch'](0);
            return _context.abrupt('return', false);

          case 9:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[0, 6]]);
  }));

  return function test(_x) {
    return _ref.apply(this, arguments);
  };
}();
// TODO: Detect base repositories?


var findRoot = function () {
  var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref2) {
    var filepath = _ref2.filepath,
        _ref2$fs = _ref2.fs,
        fs$$1 = _ref2$fs === undefined ? defaultfs() : _ref2$fs;
    var parent;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            setfs(fs$$1);
            _context2.next = 3;
            return test(filepath);

          case 3:
            if (!_context2.sent) {
              _context2.next = 5;
              break;
            }

            return _context2.abrupt('return', filepath);

          case 5:
            parent = path.dirname(filepath);

            if (!(parent === filepath)) {
              _context2.next = 8;
              break;
            }

            throw new Error('Unable to find git root');

          case 8:
            return _context2.abrupt('return', findRoot({ filepath: parent }));

          case 9:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function findRoot(_x2) {
    return _ref3.apply(this, arguments);
  };
}();

function version$1() {
  return version;
}



var commands = Object.freeze({
	add: add,
	clone: clone,
	checkout: checkout,
	commit: commit,
	fetch: fetch,
	init: init,
	list: list,
	listBranches: listBranches,
	listCommits: listCommits,
	listObjects: listObjects,
	log: log,
	pack: pack,
	push: push,
	remove: remove,
	config: config,
	unpack: unpack,
	verify: verify,
	status: status,
	findRoot: findRoot,
	version: version$1
});

function git$1(dir) {
  return dir === undefined ? new Git() : new Git().workdir(dir).gitdir(dir + '/.git');
}
git$1.commands = commands;
git$1.managers = managers;
git$1.models = models;
git$1.utils = utils;

var extend = function extend(self, array) {
  var _loop = function _loop(fn) {
    self[fn] = function (val) {
      return self.set(fn, val);
    };
  };

  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = _getIterator(array), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var fn = _step.value;

      _loop(fn);
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }
};

// This is so git().relative() is the same as git().relative(true)
// but you can still undo the flag later with git().relative(false)
var extendBool = function extendBool(self, array) {
  var _loop2 = function _loop2(fn) {
    self[fn] = function (val) {
      return self.set(fn, val !== false);
    };
  };

  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = _getIterator(array), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var fn = _step2.value;

      _loop2(fn);
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }
};

// The class is merely a fluent command/query builder
var Git = function (_Map$$1) {
  _inherits(Git, _Map$$1);

  // @constructor
  function Git() {
    var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        fs = _ref.fs,
        dir = _ref.dir,
        workdir = _ref.workdir,
        gitdir = _ref.gitdir;

    _classCallCheck(this, Git);

    var _this = _possibleConstructorReturn(this, (Git.__proto__ || _Object$getPrototypeOf(Git)).call(this));

    extend(_this, ['workdir', 'gitdir', 'remote', 'branch', 'author', 'email', 'datetime', 'depth', 'since', 'exclude', 'timestamp', 'signingKey', 'verificationKey', 'username', 'password', 'url', 'outputStream', 'inputStream', 'onprogress', 'fs']);
    extendBool(_this, ['relative']);
    if (fs) {
      _this.set('fs', fs);
      setfs(fs);
    }
    if (dir) {
      _this.set('workdir', dir);
      _this.set('gitdir', dir + '/.git');
    }
    if (workdir) {
      _this.set('workdir', workdir);
    }
    if (gitdir) {
      _this.set('gitdir', gitdir);
    }
    return _this;
  }

  _createClass(Git, [{
    key: 'version',
    value: function version$$1() {
      return version$1();
    }
    /**
     * @param {string} username
     * @param {string} password
     * @returns {Git} this
     *
     * Use with {@link #gitpush .push} and {@link #gitpull .pull} to set Basic Authentication headers.
     * This works for basic username / password auth, or the newer username / token auth
     * that is often required if 2FA is enabled.
     */

  }, {
    key: 'auth',
    value: function auth(username, password) {
      // Allow specifying it as one argument (mostly for CLI inputability)
      if (password === undefined) {
        var i = username.indexOf(':');
        if (i > -1) {
          password = username.slice(i + 1);
          username = username.slice(0, i);
        } else {
          password = ''; // Enables the .auth(GITHUB_TOKEN) no-username shorthand
        }
      }
      this.set('username', username);
      this.set('password', password);
      return this;
    }
    /**
     * @param {string} company
     * @param {string} token
     * @returns {Git} this
     *
     * Use with {@link #gitpush .push} and {@link #gitpull .pull} to set Basic Authentication headers.
     * This for is for *actual* OAuth2 tokens (not "personal access tokens").
     * Unfortunately, all the major git hosting companies have chosen different conventions!
     * Lucky for you, I already looked up and codified it for you.
     *
     * - oauth2('github', token) - Github uses `token` as the username, and 'x-oauth-basic' as the password.
     * - oauth2('bitbucket', token) - Bitbucket uses 'x-token-auth' as the username, and `token` as the password.
     * - oauth2('gitlab', token) - Gitlab uses 'oauth2' as the username, and `token` as the password.
     *
     * I will gladly accept pull requests for more companies' conventions.
     */

  }, {
    key: 'oauth2',
    value: function oauth2(company, token) {
      switch (company) {
        case 'github':
          this.set('username', token);
          this.set('password', 'x-oauth-basic');
          break;
        case 'bitbucket':
          this.set('username', 'x-token-auth');
          this.set('password', token);
          break;
        case 'gitlab':
          this.set('username', 'oauth2');
          this.set('password', token);
          break;
        default:
          throw new Error('I don\'t know how ' + company + ' expects its Basic Auth headers to be formatted for OAuth2 usage. If you do, you can use the regular \'.auth(username, password)\' to set the basic auth header yourself.');
      }
      return this;
    }
  }, {
    key: 'findRoot',
    value: function () {
      var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(dir) {
        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                return _context.abrupt('return', findRoot({
                  filepath: dir,
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function findRoot$$1(_x2) {
        return _ref2.apply(this, arguments);
      }

      return findRoot$$1;
    }()
  }, {
    key: 'init',
    value: function () {
      var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2() {
        return _regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return init({
                  gitdir: this.get('gitdir'),
                  fs: this.get('fs')
                });

              case 2:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function init$$1() {
        return _ref3.apply(this, arguments);
      }

      return init$$1;
    }()
  }, {
    key: 'fetch',
    value: function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(ref) {
        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return fetch({
                  gitdir: this.get('gitdir'),
                  ref: ref,
                  url: this.get('url'),
                  remote: this.get('remote'),
                  authUsername: this.get('username'),
                  authPassword: this.get('password'),
                  depth: this.get('depth'),
                  since: this.get('since'),
                  exclude: this.get('exclude'),
                  relative: this.get('relative'),
                  onprogress: this.get('onprogress'),
                  fs: this.get('fs')
                });

              case 2:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function fetch$$1(_x3) {
        return _ref4.apply(this, arguments);
      }

      return fetch$$1;
    }()
  }, {
    key: 'checkout',
    value: function () {
      var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee4(ref) {
        return _regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.next = 2;
                return checkout({
                  workdir: this.get('workdir'),
                  gitdir: this.get('gitdir'),
                  ref: ref,
                  remote: this.get('remote'),
                  fs: this.get('fs')
                });

              case 2:
              case 'end':
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function checkout$$1(_x4) {
        return _ref5.apply(this, arguments);
      }

      return checkout$$1;
    }()
  }, {
    key: 'clone',
    value: function () {
      var _ref6 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee5(url) {
        return _regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return clone({
                  workdir: this.get('workdir'),
                  gitdir: this.get('gitdir'),
                  url: url,
                  remote: this.get('remote'),
                  ref: this.get('branch'),
                  authUsername: this.get('username'),
                  authPassword: this.get('password'),
                  depth: this.get('depth'),
                  since: this.get('since'),
                  exclude: this.get('exclude'),
                  relative: this.get('relative'),
                  onprogress: this.get('onprogress'),
                  fs: this.get('fs')
                });

              case 2:
              case 'end':
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function clone$$1(_x5) {
        return _ref6.apply(this, arguments);
      }

      return clone$$1;
    }()
  }, {
    key: 'list',
    value: function () {
      var _ref7 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee6() {
        return _regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                return _context6.abrupt('return', list({
                  gitdir: this.get('gitdir'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function list$$1() {
        return _ref7.apply(this, arguments);
      }

      return list$$1;
    }()
  }, {
    key: 'log',
    value: function () {
      var _ref8 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee7(ref) {
        return _regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                return _context7.abrupt('return', log({
                  gitdir: this.get('gitdir'),
                  ref: ref,
                  depth: this.get('depth'),
                  since: this.get('since'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function log$$1(_x6) {
        return _ref8.apply(this, arguments);
      }

      return log$$1;
    }()
  }, {
    key: 'add',
    value: function () {
      var _ref9 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee8(filepath) {
        return _regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                return _context8.abrupt('return', add({
                  gitdir: this.get('gitdir'),
                  workdir: this.get('workdir'),
                  filepath: filepath,
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function add$$1(_x7) {
        return _ref9.apply(this, arguments);
      }

      return add$$1;
    }()
  }, {
    key: 'remove',
    value: function () {
      var _ref10 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee9(filepath) {
        return _regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                return _context9.abrupt('return', remove({
                  gitdir: this.get('gitdir'),
                  filepath: filepath,
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function remove$$1(_x8) {
        return _ref10.apply(this, arguments);
      }

      return remove$$1;
    }()
  }, {
    key: 'commit',
    value: function () {
      var _ref11 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee10(message$$1) {
        return _regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                return _context10.abrupt('return', commit({
                  gitdir: this.get('gitdir'),
                  author: {
                    name: this.get('author'),
                    email: this.get('email'),
                    timestamp: this.get('timestamp'),
                    date: this.get('datetime')
                  },
                  committer: {
                    name: this.get('author'),
                    email: this.get('email'),
                    timestamp: this.get('timestamp'),
                    date: this.get('datetime')
                  },
                  message: message$$1,
                  privateKeys: this.get('signingKey'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function commit$$1(_x9) {
        return _ref11.apply(this, arguments);
      }

      return commit$$1;
    }()
  }, {
    key: 'verify',
    value: function () {
      var _ref12 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee11(ref) {
        return _regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                return _context11.abrupt('return', verify({
                  gitdir: this.get('gitdir'),
                  publicKeys: this.get('verificationKey'),
                  ref: ref,
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context11.stop();
            }
          }
        }, _callee11, this);
      }));

      function verify$$1(_x10) {
        return _ref12.apply(this, arguments);
      }

      return verify$$1;
    }()
  }, {
    key: 'pack',
    value: function () {
      var _ref13 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee12(oids) {
        return _regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                return _context12.abrupt('return', pack({
                  gitdir: this.get('gitdir'),
                  outputStream: this.get('outputStream'),
                  oids: oids,
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context12.stop();
            }
          }
        }, _callee12, this);
      }));

      function pack$$1(_x11) {
        return _ref13.apply(this, arguments);
      }

      return pack$$1;
    }()
  }, {
    key: 'unpack',
    value: function () {
      var _ref14 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee13(oids) {
        return _regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                return _context13.abrupt('return', unpack({
                  gitdir: this.get('gitdir'),
                  inputStream: this.get('inputStream'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context13.stop();
            }
          }
        }, _callee13, this);
      }));

      function unpack$$1(_x12) {
        return _ref14.apply(this, arguments);
      }

      return unpack$$1;
    }()
  }, {
    key: 'push',
    value: function () {
      var _ref15 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee14(ref) {
        return _regeneratorRuntime.wrap(function _callee14$(_context14) {
          while (1) {
            switch (_context14.prev = _context14.next) {
              case 0:
                return _context14.abrupt('return', push({
                  gitdir: this.get('gitdir'),
                  ref: ref,
                  remote: this.get('remote'),
                  url: this.get('url'),
                  authUsername: this.get('username'),
                  authPassword: this.get('password'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context14.stop();
            }
          }
        }, _callee14, this);
      }));

      function push$$1(_x13) {
        return _ref15.apply(this, arguments);
      }

      return push$$1;
    }()
  }, {
    key: 'pull',
    value: function () {
      var _ref16 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee15(ref) {
        return _regeneratorRuntime.wrap(function _callee15$(_context15) {
          while (1) {
            switch (_context15.prev = _context15.next) {
              case 0:
                return _context15.abrupt('return', fetch({
                  gitdir: this.get('gitdir'),
                  ref: ref,
                  remote: this.get('remote'),
                  authUsername: this.get('username'),
                  authPassword: this.get('password'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context15.stop();
            }
          }
        }, _callee15, this);
      }));

      function pull(_x14) {
        return _ref16.apply(this, arguments);
      }

      return pull;
    }()
  }, {
    key: 'config',
    value: function () {
      var _ref17 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee16(path$$1, value) {
        var _args16 = arguments;
        return _regeneratorRuntime.wrap(function _callee16$(_context16) {
          while (1) {
            switch (_context16.prev = _context16.next) {
              case 0:
                if (!(_args16.length === 1)) {
                  _context16.next = 4;
                  break;
                }

                return _context16.abrupt('return', config({
                  gitdir: this.get('gitdir'),
                  path: path$$1,
                  fs: this.get('fs')
                }));

              case 4:
                return _context16.abrupt('return', config({
                  gitdir: this.get('gitdir'),
                  path: path$$1,
                  value: value,
                  fs: this.get('fs')
                }));

              case 5:
              case 'end':
                return _context16.stop();
            }
          }
        }, _callee16, this);
      }));

      function config$$1(_x15, _x16) {
        return _ref17.apply(this, arguments);
      }

      return config$$1;
    }()
  }, {
    key: 'status',
    value: function () {
      var _ref18 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee17(pathname) {
        return _regeneratorRuntime.wrap(function _callee17$(_context17) {
          while (1) {
            switch (_context17.prev = _context17.next) {
              case 0:
                return _context17.abrupt('return', status({
                  gitdir: this.get('gitdir'),
                  workdir: this.get('workdir'),
                  pathname: pathname,
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context17.stop();
            }
          }
        }, _callee17, this);
      }));

      function status$$1(_x17) {
        return _ref18.apply(this, arguments);
      }

      return status$$1;
    }()
  }, {
    key: 'listBranches',
    value: function () {
      var _ref19 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee18() {
        return _regeneratorRuntime.wrap(function _callee18$(_context18) {
          while (1) {
            switch (_context18.prev = _context18.next) {
              case 0:
                return _context18.abrupt('return', listBranches({
                  gitdir: this.get('gitdir'),
                  fs: this.get('fs')
                }));

              case 1:
              case 'end':
                return _context18.stop();
            }
          }
        }, _callee18, this);
      }));

      function listBranches$$1() {
        return _ref19.apply(this, arguments);
      }

      return listBranches$$1;
    }()
  }]);

  return Git;
}(_Map);

git$1.Git = Git;

module.exports = git$1;
//# sourceMappingURL=index-umd.js.map
